#!/usr/bin/env python
"""Provides conversion functions to be used during RDFProtoStruct migration."""

from google.protobuf import any_pb2
from grr_response_core.lib.rdfvalues import structs as rdf_structs
from grr_response_proto import semantic_pb2


def ToProtoSemanticDescriptor(
    rdf: rdf_structs.SemanticDescriptor,
) -> semantic_pb2.SemanticDescriptor:
  return rdf.AsPrimitiveProto()


def ToRDFSemanticDescriptor(
    proto: semantic_pb2.SemanticDescriptor,
) -> rdf_structs.SemanticDescriptor:
  return rdf_structs.SemanticDescriptor.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoAny(rdf: rdf_structs.AnyValue) -> any_pb2.Any:
  return rdf.AsPrimitiveProto()


def ToRDFAnyValue(proto: any_pb2.Any) -> rdf_structs.AnyValue:
  return rdf_structs.AnyValue.FromSerializedBytes(proto.SerializeToString())



#!/usr/bin/env python
"""Provides conversion functions to be used during RDFProtoStruct migration."""

from grr_response_core.lib.rdfvalues import client_action as rdf_client_action
from grr_response_proto import flows_pb2
from grr_response_proto import jobs_pb2


def ToProtoEchoRequest(
    rdf: rdf_client_action.EchoRequest,
) -> jobs_pb2.EchoRequest:
  return rdf.AsPrimitiveProto()


def ToRDFEchoRequest(
    proto: jobs_pb2.EchoRequest,
) -> rdf_client_action.EchoRequest:
  return rdf_client_action.EchoRequest.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoExecuteBinaryRequest(
    rdf: rdf_client_action.ExecuteBinaryRequest,
) -> jobs_pb2.ExecuteBinaryRequest:
  return rdf.AsPrimitiveProto()


def ToRDFExecuteBinaryRequest(
    proto: jobs_pb2.ExecuteBinaryRequest,
) -> rdf_client_action.ExecuteBinaryRequest:
  return rdf_client_action.ExecuteBinaryRequest.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoExecuteBinaryResponse(
    rdf: rdf_client_action.ExecuteBinaryResponse,
) -> jobs_pb2.ExecuteBinaryResponse:
  return rdf.AsPrimitiveProto()


def ToRDFExecuteBinaryResponse(
    proto: jobs_pb2.ExecuteBinaryResponse,
) -> rdf_client_action.ExecuteBinaryResponse:
  return rdf_client_action.ExecuteBinaryResponse.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoExecutePythonRequest(
    rdf: rdf_client_action.ExecutePythonRequest,
) -> jobs_pb2.ExecutePythonRequest:
  return rdf.AsPrimitiveProto()


def ToRDFExecutePythonRequest(
    proto: jobs_pb2.ExecutePythonRequest,
) -> rdf_client_action.ExecutePythonRequest:
  return rdf_client_action.ExecutePythonRequest.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoExecutePythonResponse(
    rdf: rdf_client_action.ExecutePythonResponse,
) -> jobs_pb2.ExecutePythonResponse:
  return rdf.AsPrimitiveProto()


def ToRDFExecutePythonResponse(
    proto: jobs_pb2.ExecutePythonResponse,
) -> rdf_client_action.ExecutePythonResponse:
  return rdf_client_action.ExecutePythonResponse.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoExecuteRequest(
    rdf: rdf_client_action.ExecuteRequest,
) -> jobs_pb2.ExecuteRequest:
  return rdf.AsPrimitiveProto()


def ToRDFExecuteRequest(
    proto: jobs_pb2.ExecuteRequest,
) -> rdf_client_action.ExecuteRequest:
  return rdf_client_action.ExecuteRequest.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoExecuteResponse(
    rdf: rdf_client_action.ExecuteResponse,
) -> jobs_pb2.ExecuteResponse:
  return rdf.AsPrimitiveProto()


def ToRDFExecuteResponse(
    proto: jobs_pb2.ExecuteResponse,
) -> rdf_client_action.ExecuteResponse:
  return rdf_client_action.ExecuteResponse.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoIterator(rdf: rdf_client_action.Iterator) -> jobs_pb2.Iterator:
  return rdf.AsPrimitiveProto()


def ToRDFIterator(proto: jobs_pb2.Iterator) -> rdf_client_action.Iterator:
  return rdf_client_action.Iterator.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoListDirRequest(
    rdf: rdf_client_action.ListDirRequest,
) -> jobs_pb2.ListDirRequest:
  return rdf.AsPrimitiveProto()


def ToRDFListDirRequest(
    proto: jobs_pb2.ListDirRequest,
) -> rdf_client_action.ListDirRequest:
  return rdf_client_action.ListDirRequest.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoGetFileStatRequest(
    rdf: rdf_client_action.GetFileStatRequest,
) -> jobs_pb2.GetFileStatRequest:
  return rdf.AsPrimitiveProto()


def ToRDFGetFileStatRequest(
    proto: jobs_pb2.GetFileStatRequest,
) -> rdf_client_action.GetFileStatRequest:
  return rdf_client_action.GetFileStatRequest.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoFingerprintTuple(
    rdf: rdf_client_action.FingerprintTuple,
) -> jobs_pb2.FingerprintTuple:
  return rdf.AsPrimitiveProto()


def ToRDFFingerprintTuple(
    proto: jobs_pb2.FingerprintTuple,
) -> rdf_client_action.FingerprintTuple:
  return rdf_client_action.FingerprintTuple.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoFingerprintRequest(
    rdf: rdf_client_action.FingerprintRequest,
) -> jobs_pb2.FingerprintRequest:
  return rdf.AsPrimitiveProto()


def ToRDFFingerprintRequest(
    proto: jobs_pb2.FingerprintRequest,
) -> rdf_client_action.FingerprintRequest:
  return rdf_client_action.FingerprintRequest.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoFingerprintResponse(
    rdf: rdf_client_action.FingerprintResponse,
) -> jobs_pb2.FingerprintResponse:
  return rdf.AsPrimitiveProto()


def ToRDFFingerprintResponse(
    proto: jobs_pb2.FingerprintResponse,
) -> rdf_client_action.FingerprintResponse:
  return rdf_client_action.FingerprintResponse.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoWMIRequest(rdf: rdf_client_action.WMIRequest) -> jobs_pb2.WMIRequest:
  return rdf.AsPrimitiveProto()


def ToRDFWMIRequest(proto: jobs_pb2.WMIRequest) -> rdf_client_action.WMIRequest:
  return rdf_client_action.WMIRequest.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoStatFSRequest(
    rdf: rdf_client_action.StatFSRequest,
) -> jobs_pb2.StatFSRequest:
  return rdf.AsPrimitiveProto()


def ToRDFStatFSRequest(
    proto: jobs_pb2.StatFSRequest,
) -> rdf_client_action.StatFSRequest:
  return rdf_client_action.StatFSRequest.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoListNetworkConnectionsArgs(
    rdf: rdf_client_action.ListNetworkConnectionsArgs,
) -> flows_pb2.ListNetworkConnectionsArgs:
  return rdf.AsPrimitiveProto()


def ToRDFListNetworkConnectionsArgs(
    proto: flows_pb2.ListNetworkConnectionsArgs,
) -> rdf_client_action.ListNetworkConnectionsArgs:
  return rdf_client_action.ListNetworkConnectionsArgs.FromSerializedBytes(
      proto.SerializeToString()
  )



#!/usr/bin/env python
# Copyright 2011 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This file was imported with permission from
# http://verify-sigs.googlecode.com/files/verify-sigs-1.1.tar.bz2
"""Fingerprinter class and some utility functions to exercise it.

While this file contains a main and some top-level functions, those
are meant for exploration and debugging. Intended use is through the
Fingerprinter, as exemplified in main.
"""


import collections
import hashlib
import os
import struct


# pylint: disable=g-bad-name
# Two classes given named tupes for ranges and relative ranges.
Range = collections.namedtuple('Range', 'start end')
RelRange = collections.namedtuple('RelRange', 'start len')

# pylint: enable=g-bad-name


class Finger:
  """A Finger defines how to hash a file to get specific fingerprints.

  The Finger contains one or more hash functions, a set of ranges in the
  file that are to be processed with these hash functions, and relevant
  metadata and accessor methods.

  While one Finger provides potentially multiple hashers, they all get
  fed the same ranges of the file.
  """

  def __init__(self, hashers, ranges, metadata_dict):
    self.hashers = hashers
    self.ranges = ranges
    self.metadata = metadata_dict

  def CurrentRange(self):
    """The working range of this Finger. Returns None if there is none."""
    if self.ranges:
      return self.ranges[0]
    return None

  def ConsumeRange(self, start, end):
    """Consumes an entire range, or part thereof.

    If the finger has no ranges left, or the current range start is higher
    than the end of the consumed block, nothing happens. Otherwise,
    the current range is adjusted for the consumed block, or removed,
    if the entire block is consumed. For things to work, the consumed
    range and the current finger starts must be equal, and the length
    of the consumed range may not exceed the length of the current range.

    Args:
      start: Beginning of range to be consumed.
      end: First offset after the consumed range (end + 1).

    Raises:
      RuntimeError: if the start position of the consumed range is
          higher than the start of the current range in the finger, or if
          the consumed range cuts across block boundaries.
    """
    old = self.CurrentRange()
    if old is None:
      return
    if old.start > start:
      if old.start < end:
        raise RuntimeError('Block end too high.')
      return
    if old.start < start:
      raise RuntimeError('Block start too high.')
    if old.end == end:
      del self.ranges[0]
    elif old.end > end:
      self.ranges[0] = Range(end, old.end)
    else:
      raise RuntimeError('Block length exceeds range.')

  def HashBlock(self, block):
    """Given a data block, feed it to all the registered hashers."""
    for hasher in self.hashers:
      hasher.update(block)


class Fingerprinter:
  """Compute different types of cryptographic hashes over a file.

  Depending on type of file and mode of invocation, filetype-specific or
  generic hashes get computed over a file. Different hashes can cover
  different ranges of the file. The file is read only once. Memory
  use of class objects is dominated by min(file size, block size),
  as defined below.

  The class delivers an array with dicts of hashes by file type. Where
  appropriate, embedded signature data is also returned from the file.

  Suggested use:
  - Provide file object at initialisation time.
  - Invoke one or more of the Eval* functions, with your choice of hashers.
  - Call HashIt and take from the resulting dict what you need.
  """

  BLOCK_SIZE = 1000000
  GENERIC_HASH_CLASSES = (
      hashlib.md5,
      hashlib.sha1,
      hashlib.sha256,
      hashlib.sha512,
  )
  AUTHENTICODE_HASH_CLASSES = (hashlib.md5, hashlib.sha1)

  def __init__(self, file_obj):
    self.fingers = []
    self.file = file_obj
    self.file.seek(0, os.SEEK_END)
    self.filelength = self.file.tell()

  def _GetNextInterval(self):
    """Returns the next Range of the file that is to be hashed.

    For all fingers, inspect their next expected range, and return the
    lowest uninterrupted range of interest. If the range is larger than
    BLOCK_SIZE, truncate it.

    Returns:
      Next range of interest in a Range namedtuple.
    """
    ranges = [x.CurrentRange() for x in self.fingers]
    starts = set([r.start for r in ranges if r])
    ends = set([r.end for r in ranges if r])
    if not starts:
      return None
    min_start = min(starts)
    starts.remove(min_start)
    ends |= starts
    min_end = min(ends)
    if min_end - min_start > self.BLOCK_SIZE:
      min_end = min_start + self.BLOCK_SIZE
    return Range(min_start, min_end)

  def _AdjustIntervals(self, start, end):
    for finger in self.fingers:
      finger.ConsumeRange(start, end)

  def _HashBlock(self, block, start, end):
    """_HashBlock feeds data blocks into the hashers of fingers.

    This function must be called before adjusting fingers for next
    interval, otherwise the lack of remaining ranges will cause the
    block not to be hashed for a specific finger.

    Start and end are used to validate the expected ranges, to catch
    unexpected use of that logic.

    Args:
      block: The data block.
      start: Beginning offset of this block.
      end: Offset of the next byte after the block.

    Raises:
      RuntimeError: If the provided and expected ranges don't match.
    """
    for finger in self.fingers:
      expected_range = finger.CurrentRange()
      if expected_range is None:
        continue
      if (
          start > expected_range.start
          or (start == expected_range.start and end > expected_range.end)
          or (start < expected_range.start and end > expected_range.start)
      ):
        raise RuntimeError('Cutting across fingers.')
      if start == expected_range.start:
        finger.HashBlock(block)

  def HashIt(self):
    """Finalizing function for the Fingerprint class.

    This method applies all the different hash functions over the
    previously specified different ranges of the input file, and
    computes the resulting hashes.

    After calling HashIt, the state of the object is reset to its
    initial state, with no fingers defined.

    Returns:
      An array of dicts, with each dict containing name of fingerprint
      type, names of hashes and values, and additional, type-dependent
      key / value pairs, such as an array of SignedData tuples for the
      PE/COFF fingerprint type.

    Raises:
       RuntimeError: when internal inconsistencies occur.
    """
    while True:
      interval = self._GetNextInterval()
      if interval is None:
        break
      self.file.seek(interval.start, os.SEEK_SET)
      block = self.file.read(interval.end - interval.start)
      if len(block) != interval.end - interval.start:
        raise RuntimeError('Short read on file.')
      self._HashBlock(block, interval.start, interval.end)
      self._AdjustIntervals(interval.start, interval.end)

    results = []
    for finger in self.fingers:
      res = {}
      leftover = finger.CurrentRange()
      if leftover:
        if (
            len(finger.ranges) > 1
            or leftover.start != self.filelength
            or leftover.end != self.filelength
        ):
          raise RuntimeError('Non-empty range remains.')
      res.update(finger.metadata)
      for hasher in finger.hashers:
        res[str(hasher.name)] = hasher.digest()
      results.append(res)

    # Clean out things for a fresh start (on the same file object).
    self.fingers = []

    # Make sure the results come back in 'standard' order, regardless of the
    # order in which fingers were added. Helps with reproducing test results.
    return sorted(results, key=lambda r: r['name'])

  def EvalGeneric(self, hashers=None):
    """Causes the entire file to be hashed by the given hash functions.

    This sets up a 'finger' for fingerprinting, where the entire file
    is passed through a pre-defined (or user defined) set of hash functions.

    Args:
      hashers: An iterable of hash classes (e.g. out of hashlib) which will be
        instantiated for use. If hashers is not provided, or is provided as
        'None', the default hashers will get used. To invoke this without
        hashers, provide an empty list.

    Returns:
      Always True, as all files are 'generic' files.
    """
    if hashers is None:
      hashers = Fingerprinter.GENERIC_HASH_CLASSES
    hashfuncs = [x() for x in hashers]
    finger = Finger(hashfuncs, [Range(0, self.filelength)], {'name': 'generic'})
    self.fingers.append(finger)
    return True

  def _PecoffHeaderParser(self):
    """Parses PECOFF headers.

    Reads header magic and some data structures in a file to determine if
    it is a valid PECOFF header, and figure out the offsets at which
    relevant data is stored.
    While this code contains multiple seeks and small reads, that is
    compensated by the underlying libc buffering mechanism.

    Returns:
      None if the parsed file is not PECOFF.
      A dict with offsets and lengths for CheckSum, CertTable, and SignedData
      fields in the PECOFF binary, for those that are present.
    """
    extents = {}
    self.file.seek(0, os.SEEK_SET)
    buf = self.file.read(2)
    if buf != b'MZ':
      return None
    self.file.seek(0x3C, os.SEEK_SET)
    buf = self.file.read(4)
    pecoff_sig_offset = struct.unpack('<I', buf)[0]
    if pecoff_sig_offset >= self.filelength:
      return None
    self.file.seek(pecoff_sig_offset, os.SEEK_SET)
    buf = self.file.read(4)
    if buf != b'PE\0\0':
      return None
    self.file.seek(pecoff_sig_offset + 20, os.SEEK_SET)
    buf = self.file.read(2)
    optional_header_size = struct.unpack('<H', buf)[0]
    optional_header_offset = pecoff_sig_offset + 4 + 20
    if optional_header_size + optional_header_offset > self.filelength:
      # This is not strictly a failure for windows, but such files better
      # be treated as generic files. They can not be carrying SignedData.
      return None
    if optional_header_size < 68:
      # We can't do authenticode-style hashing. If this is a valid binary,
      # which it can be, the header still does not even contain a checksum.
      return None
    self.file.seek(optional_header_offset, os.SEEK_SET)
    buf = self.file.read(2)
    image_magic = struct.unpack('<H', buf)[0]
    if image_magic == 0x10B:
      # 32 bit
      rva_base = optional_header_offset + 92
      cert_base = optional_header_offset + 128
    elif image_magic == 0x20B:
      # 64 bit
      rva_base = optional_header_offset + 108
      cert_base = optional_header_offset + 144
    else:
      # A ROM image or such, not in the PE/COFF specs. Not sure what to do.
      return None
    extents['CheckSum'] = RelRange(optional_header_offset + 64, 4)
    self.file.seek(rva_base, os.SEEK_SET)
    buf = self.file.read(4)
    number_of_rva = struct.unpack('<I', buf)[0]
    if (
        number_of_rva < 5
        or optional_header_offset + optional_header_size < cert_base + 8
    ):
      return extents
    extents['CertTable'] = RelRange(cert_base, 8)

    self.file.seek(cert_base, os.SEEK_SET)
    buf = self.file.read(8)
    start, length = struct.unpack('<II', buf)
    if (
        length == 0
        or start < optional_header_offset + optional_header_size
        or start + length > self.filelength
    ):
      # The location of the SignedData blob is just wrong (or there is none).
      # Ignore it -- everything else we did still makes sense.
      return extents
    extents['SignedData'] = RelRange(start, length)
    return extents

  def _CollectSignedData(self, extent):
    """Extracts signedData blob from PECOFF binary and parses first layer."""
    start, length = extent

    self.file.seek(start, os.SEEK_SET)
    buf = self.file.read(length)
    signed_data = []
    # This loop ignores trailing cruft, or too-short signedData chunks.
    while len(buf) >= 8:
      dw_length, w_revision, w_cert_type = struct.unpack('<IHH', buf[:8])
      if dw_length < 8:
        # If the entire blob is smaller than its header, bail out.
        return signed_data
      b_cert = buf[8:dw_length]
      buf = buf[(dw_length + 7) & 0x7FFFFFF8 :]
      signed_data.append((w_revision, w_cert_type, b_cert))
    return signed_data

  def EvalPecoff(self, hashers=None):
    """If the file is a PE/COFF file, computes authenticode hashes on it.

    This checks if the input file is a valid PE/COFF image file (e.g. a
    Windows binary, driver, or DLL) and if yes, sets up a 'finger' for
    fingerprinting in Authenticode style.
    If available, the 'SignedData' section of the image file is retrieved,
    and parsed into its constituent parts. An array of tuples of these
    parts is added to results by HashIt()

    Args:
      hashers: An iterable of hash classes (e.g. out of hashlib) which will be
        instantiated for use. If 'None' is provided, a default set of hashers is
        used. To select no hash function (e.g. to only extract metadata), use an
        empty iterable.

    Returns:
      True if the file is detected as a valid PE/COFF image file,
      False otherwise.
    """
    try:
      extents = self._PecoffHeaderParser()
    except struct.error:
      # Parsing the header failed. Just ignore this, and claim
      # that the file is not a valid PE/COFF image file.
      extents = None
    if extents is None:
      return False

    signed_data = None
    ranges = []
    start = 0
    # Ordering of these conditions matches expected order in file.
    # If a condition holds true, the matching range is skipped for hashing.
    if 'CheckSum' in extents:
      ranges.append(Range(start, end=extents['CheckSum'].start))
      start = sum(extents['CheckSum'])
      # New start now points past CheckSum area.
    if 'CertTable' in extents:
      ranges.append(Range(start, end=extents['CertTable'].start))
      start = sum(extents['CertTable'])
      # New start now points past CertTable area.
    if 'SignedData' in extents:
      # Exclude the range even if the blob itself can't be parsed correctly.
      ranges.append(Range(start, end=extents['SignedData'].start))
      start = sum(extents['SignedData'])
      # New start now points past SignedData area.
      signed_data = self._CollectSignedData(extents['SignedData'])
    ranges.append(Range(start, end=self.filelength))

    if hashers is None:
      hashers = Fingerprinter.AUTHENTICODE_HASH_CLASSES
    hashfuncs = [x() for x in hashers]
    metadata = {'name': 'pecoff'}
    if signed_data:
      metadata['SignedData'] = signed_data
    finger = Finger(hashfuncs, ranges, metadata)
    self.fingers.append(finger)
    return True



#!/usr/bin/env python
"""Provides conversion functions to be used during RDFProtoStruct migration."""

from grr_response_core.lib.rdfvalues import memory as rdf_memory
from grr_response_proto import flows_pb2


def ToProtoYaraSignatureShard(
    rdf: rdf_memory.YaraSignatureShard,
) -> flows_pb2.YaraSignatureShard:
  return rdf.AsPrimitiveProto()


def ToRDFYaraSignatureShard(
    proto: flows_pb2.YaraSignatureShard,
) -> rdf_memory.YaraSignatureShard:
  return rdf_memory.YaraSignatureShard.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoYaraProcessScanRequest(
    rdf: rdf_memory.YaraProcessScanRequest,
) -> flows_pb2.YaraProcessScanRequest:
  return rdf.AsPrimitiveProto()


def ToRDFYaraProcessScanRequest(
    proto: flows_pb2.YaraProcessScanRequest,
) -> rdf_memory.YaraProcessScanRequest:
  return rdf_memory.YaraProcessScanRequest.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoProcessMemoryError(
    rdf: rdf_memory.ProcessMemoryError,
) -> flows_pb2.ProcessMemoryError:
  return rdf.AsPrimitiveProto()


def ToRDFProcessMemoryError(
    proto: flows_pb2.ProcessMemoryError,
) -> rdf_memory.ProcessMemoryError:
  return rdf_memory.ProcessMemoryError.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoYaraStringMatch(
    rdf: rdf_memory.YaraStringMatch,
) -> flows_pb2.YaraStringMatch:
  return rdf.AsPrimitiveProto()


def ToRDFYaraStringMatch(
    proto: flows_pb2.YaraStringMatch,
) -> rdf_memory.YaraStringMatch:
  return rdf_memory.YaraStringMatch.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoYaraMatch(rdf: rdf_memory.YaraMatch) -> flows_pb2.YaraMatch:
  return rdf.AsPrimitiveProto()


def ToRDFYaraMatch(proto: flows_pb2.YaraMatch) -> rdf_memory.YaraMatch:
  return rdf_memory.YaraMatch.FromSerializedBytes(proto.SerializeToString())


def ToProtoYaraProcessScanMatch(
    rdf: rdf_memory.YaraProcessScanMatch,
) -> flows_pb2.YaraProcessScanMatch:
  return rdf.AsPrimitiveProto()


def ToRDFYaraProcessScanMatch(
    proto: flows_pb2.YaraProcessScanMatch,
) -> rdf_memory.YaraProcessScanMatch:
  return rdf_memory.YaraProcessScanMatch.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoYaraProcessScanMiss(
    rdf: rdf_memory.YaraProcessScanMiss,
) -> flows_pb2.YaraProcessScanMiss:
  return rdf.AsPrimitiveProto()


def ToRDFYaraProcessScanMiss(
    proto: flows_pb2.YaraProcessScanMiss,
) -> rdf_memory.YaraProcessScanMiss:
  return rdf_memory.YaraProcessScanMiss.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoYaraProcessScanResponse(
    rdf: rdf_memory.YaraProcessScanResponse,
) -> flows_pb2.YaraProcessScanResponse:
  return rdf.AsPrimitiveProto()


def ToRDFYaraProcessScanResponse(
    proto: flows_pb2.YaraProcessScanResponse,
) -> rdf_memory.YaraProcessScanResponse:
  return rdf_memory.YaraProcessScanResponse.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoYaraProcessDumpArgs(
    rdf: rdf_memory.YaraProcessDumpArgs,
) -> flows_pb2.YaraProcessDumpArgs:
  return rdf.AsPrimitiveProto()


def ToRDFYaraProcessDumpArgs(
    proto: flows_pb2.YaraProcessDumpArgs,
) -> rdf_memory.YaraProcessDumpArgs:
  return rdf_memory.YaraProcessDumpArgs.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoProcessMemoryRegion(
    rdf: rdf_memory.ProcessMemoryRegion,
) -> flows_pb2.ProcessMemoryRegion:
  return rdf.AsPrimitiveProto()


def ToRDFProcessMemoryRegion(
    proto: flows_pb2.ProcessMemoryRegion,
) -> rdf_memory.ProcessMemoryRegion:
  return rdf_memory.ProcessMemoryRegion.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoYaraProcessDumpInformation(
    rdf: rdf_memory.YaraProcessDumpInformation,
) -> flows_pb2.YaraProcessDumpInformation:
  return rdf.AsPrimitiveProto()


def ToRDFYaraProcessDumpInformation(
    proto: flows_pb2.YaraProcessDumpInformation,
) -> rdf_memory.YaraProcessDumpInformation:
  return rdf_memory.YaraProcessDumpInformation.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoYaraProcessDumpResponse(
    rdf: rdf_memory.YaraProcessDumpResponse,
) -> flows_pb2.YaraProcessDumpResponse:
  return rdf.AsPrimitiveProto()


def ToRDFYaraProcessDumpResponse(
    proto: flows_pb2.YaraProcessDumpResponse,
) -> rdf_memory.YaraProcessDumpResponse:
  return rdf_memory.YaraProcessDumpResponse.FromSerializedBytes(
      proto.SerializeToString()
  )



#!/usr/bin/env python
"""The path detection interface (base) class definitions."""

import abc
import shlex


def SplitIntoComponents(str_in):
  """Splits strings into space-separated components.

  The difference between SplitIntoComponents and .split(" ") is that
  the former tries to respect single and double quotes and strip them
  using lexical rules typically used in command line interpreters.

  Args:
    str_in: Source string to be split into components.

  Returns:
    str_in represented as a list of components. Components is effectively a
    space-separated representation of a source string. " ".join(components)
    should produce the original string (with pairs of single/double quotes
    stripped).
  """

  if str_in.startswith(('"', "'")):
    return shlex.split(str_in)
  else:
    components = str_in.split(" ", 1)
    if len(components) > 1:
      return [components[0]] + SplitIntoComponents(components[1])
    else:
      return components


class Extractor(metaclass=abc.ABCMeta):
  """Base class for paths extractors."""

  @abc.abstractmethod
  def Extract(self, components):
    """Extracts interesting paths from a given path.

    Args:
      components: Source string represented as a list of components. Components
        are generated by applying SplitIntoComponents to a string. Components is
        effectively a space-separated representation of a source string. "
        ".join(components) should produce the original string (with pairs of
        single/double quotes stripped). See SplitIntoComponents for details.

    Returns:
      A list of extracted paths (as strings).
    """
    raise NotImplementedError()


class PostProcessor(metaclass=abc.ABCMeta):
  """Base class for paths post-processors."""

  @abc.abstractmethod
  def Process(self, path):
    """Post-processes given detected path.

    Args:
      path: Path (as a string) to post-process.

    Returns:
      A list of processed paths.
    """
    raise NotImplementedError()


class Detector:
  """Configurable class that implements all detection steps."""

  def __init__(self, extractors=None, post_processors=None):
    """Detector constructor.

    Detector is configured with a list of extractors and a list of post
    processors. When Detect() is called, Detector will first split the
    source string into components using SplitIntoComponents function,
    then it will feed the resulting components to extractors. Whatever
    the extractors produce will be fed to post processors and then
    returned to the user.

    Args:
      extractors: Extractors to be applied to source strings.
      post_processors: Post processors to be applied to source strings.
    """
    self.extractors = extractors or []
    self.post_processors = post_processors or []

  def Detect(self, str_in):
    """Detects paths in a given string.

    Args:
      str_in: String where the paths should be detected.

    Returns:
      A list of paths (as strings) detected inside the given string.
    """

    components = SplitIntoComponents(str_in)

    extracted_paths = set()
    for extractor in self.extractors:
      extracted_paths.update(extractor.Extract(components))

    results = set(extracted_paths)

    for post_processor in self.post_processors:
      processed_results = set()
      for result in results:
        processed_results.update(post_processor.Process(result))
      results = processed_results

    return results



#!/usr/bin/env python
"""(De-)serialization to bytes, wire format, and human readable strings."""

import abc

from grr_response_core.lib import rdfvalue
from grr_response_core.lib.util import precondition


class Converter(metaclass=abc.ABCMeta):
  """Interface for (de-)serializing types to bytes, wire format, and strings."""

  @abc.abstractproperty
  def protobuf_type(self):
    pass

  @abc.abstractmethod
  def FromBytes(self, value: bytes):
    """Deserializes a value from bytes outputted by ToBytes."""

  @abc.abstractmethod
  def ToBytes(self, value) -> bytes:
    """Serializes `value` into bytes which can be parsed with FromBytes."""

  @abc.abstractmethod
  def FromWireFormat(self, value):
    """Deserializes a value from a primitive outputted by ToWireFormat."""

  @abc.abstractmethod
  def ToWireFormat(self, value):
    """Serializes to a primitive which can be parsed with FromWireFormat."""

  @abc.abstractmethod
  def FromHumanReadable(self, string: str):
    """Deserializes a value from a string outputted by str(value)."""


class BoolConverter(Converter):
  """Converter for Python's `bool`."""

  protobuf_type = "unsigned_integer"
  wrapping_type = bool

  def FromBytes(self, value: bytes) -> bool:
    return bool(int(value))

  def ToBytes(self, value: bool) -> bytes:
    return b"1" if value else b"0"

  def FromWireFormat(self, value: int) -> bool:
    precondition.AssertType(value, int)
    return bool(value)

  def ToWireFormat(self, value: bool) -> int:
    return 1 if value else 0

  def FromHumanReadable(self, string: str) -> bool:
    upper_string = string.upper()
    if upper_string == "TRUE" or string == "1":
      return True
    elif upper_string == "FALSE" or string == "0":
      return False
    else:
      raise ValueError("Unparsable boolean string: `%s`" % string)


class RDFValueConverter(Converter):
  """Converter for rdfvalue.RDFValue."""

  def __init__(self, cls):
    super().__init__()
    self._cls = cls

  @property
  def protobuf_type(self):
    return self._cls.protobuf_type

  def FromBytes(self, value: bytes):
    return self._cls.FromSerializedBytes(value)

  def ToBytes(self, value: rdfvalue.RDFValue) -> bytes:
    precondition.AssertType(value, self._cls)
    return value.SerializeToBytes()

  def FromWireFormat(self, value):
    return self._cls.FromWireFormat(value)

  def ToWireFormat(self, value: rdfvalue.RDFValue):
    precondition.AssertType(value, self._cls)
    return value.SerializeToWireFormat()

  def FromHumanReadable(self, string: str) -> rdfvalue.RDFValue:
    if issubclass(self._cls, rdfvalue.RDFPrimitive):
      return self._cls.FromHumanReadable(string)
    else:
      raise ValueError()


def _GetFactory(cls):
  precondition.AssertType(cls, type)
  if issubclass(cls, rdfvalue.RDFValue):
    return RDFValueConverter(cls)
  elif cls is bool:
    return BoolConverter()
  raise ValueError("Unknown class {}".format(cls))


def GetProtobufType(cls):
  """Returns the protobuf type required by structs.py for the given cls."""
  return _GetFactory(cls).protobuf_type


def FromHumanReadable(cls, string: str):
  """Deserializes a value of `cls` from a string outputted by str(value)."""
  precondition.AssertType(string, str)
  return _GetFactory(cls).FromHumanReadable(string)


def ToWireFormat(value):
  """Serializes to a primitive which can be parsed with FromWireFormat."""
  return _GetFactory(type(value)).ToWireFormat(value)


def FromWireFormat(cls, value):
  """Deserializes a value of cls from a primitive outputted by ToWireFormat."""
  return _GetFactory(cls).FromWireFormat(value)


def FromBytes(cls, value: bytes):
  """Deserializes a value of `cls` from bytes outputted by ToBytes."""
  precondition.AssertType(value, bytes)
  return _GetFactory(cls).FromBytes(value)


def ToBytes(value) -> bytes:
  """Serializes `value` into bytes which can be parsed with FromBytes."""
  return _GetFactory(type(value)).ToBytes(value)



#!/usr/bin/env python
"""Tests for utility classes."""

import contextlib
import io
import os
import threading
from unittest import mock
import zipfile

from absl import app
from absl.testing import absltest

from grr_response_core.lib import utils
from grr.test_lib import test_lib

# Test method names don't conform with Google style
# pylint: disable=g-bad-name


class StoreTests(test_lib.GRRBaseTest):
  """Store tests."""

  def test01StoreExpiration(self):
    """Testing store removes objects when full."""
    s = utils.FastStore(max_size=5)
    keys = []
    for i in range(0, 100):
      keys.append(s.Put(i, i))

    # This should not raise
    s.Get(keys[-1])

    # This should raise though
    self.assertRaises(KeyError, s.Get, keys[0])

  def test02StoreRefresh(self):
    """Test that store keeps recently gotten objects fresh."""
    s = utils.FastStore(max_size=5)
    keys = []
    for i in range(0, 5):
      keys.append(s.Put(i, i))

    # This should not raise because keys[0] should be refreshed each time its
    # gotten
    for i in range(0, 1000):
      s.Get(keys[0])
      s.Put(i, i)

  def test03Expire(self):
    """Tests the expire mechanism."""
    s = utils.FastStore(max_size=100)
    key = "test1"
    s.Put(key, 1)

    # This should not raise
    self.assertEqual(s.Get(key), 1)
    s.ExpireObject(key)

    self.assertRaises(KeyError, s.Get, key)

  def test04KillObject(self):
    """Test that callbacks are called using object destruction."""
    results = []

    class TestStore(utils.FastStore):

      def KillObject(self, obj):
        results.append(obj)

    s = TestStore(max_size=5)
    for i in range(0, 10):
      s.Put(i, i)

    # Only the first 5 messages have been expired (and hence called)
    self.assertEqual(results, list(range(0, 5)))

  def test05TimeBasedCache(self):

    key = "key"
    tested_cache = utils.TimeBasedCache(max_age=50)
    with test_lib.FakeTime(100):

      # Stop the housekeeper thread - we test it explicitly here
      tested_cache.exit = True
      tested_cache.Put(key, "hello")

      self.assertEqual(tested_cache.Get(key), "hello")

    with test_lib.FakeTime(160):

      # Force the housekeeper to run
      tested_cache.house_keeper_thread.target()

      # This should now be expired
      self.assertRaises(KeyError, tested_cache.Get, key)

  def testTimeBasedCacheSingleThread(self):

    utils.TimeBasedCache()
    num_threads = threading.active_count()
    utils.TimeBasedCache()
    self.assertEqual(threading.active_count(), num_threads)

  def testWeakRefSet(self):

    c1 = utils.TimeBasedCache()
    c2 = utils.TimeBasedCache()

    self.assertIn(c1, utils.TimeBasedCache.active_caches)
    self.assertIn(c2, utils.TimeBasedCache.active_caches)

    l = len(utils.TimeBasedCache.active_caches)

    del c1

    # This should work even though the weak ref to c1 should be gone.
    utils.TimeBasedCache.house_keeper_thread.target()

    # Make sure it's actually gone.
    self.assertLess(len(utils.TimeBasedCache.active_caches), l)


class UtilsTest(test_lib.GRRBaseTest):
  """Utilities tests."""

  def testNormpath(self):
    """Test our Normpath."""
    data = [
        ("foo/../../../../", "/"),
        ("/foo/../../../../bar", "/bar"),
        ("/foo/bar/../3sdfdfsa/.", "/foo/3sdfdfsa"),
        ("../foo/bar", "/foo/bar"),
        ("./foo/bar", "/foo/bar"),
        ("/", "/"),
    ]

    for test, expected in data:
      self.assertEqual(expected, utils.NormalizePath(test))

  def FormatAsHexStringTest(self):
    self.assertEqual(utils.FormatAsHexString(10), "0x1b")
    self.assertEqual(utils.FormatAsHexString(10, 4), "0x001b")
    self.assertEqual(utils.FormatAsHexString(10, 16), "0x000000000000001b")
    # No trailing "L".
    self.assertEqual(utils.FormatAsHexString(int(1e19)), "0x8ac7230489e80000")
    self.assertEqual(
        utils.FormatAsHexString(int(1e19), 5), "0x8ac7230489e80000"
    )

  def testXor(self):
    test_str = b"foobar4815162342"
    for key in [1, 5, 123, 255]:
      xor_str = utils.Xor(test_str, key)
      self.assertNotEqual(xor_str, test_str)
      xor_str = utils.Xor(xor_str, key)
      self.assertEqual(xor_str, test_str)

  def LinkedListTest(self):

    l = utils.LinkedList()

    # Test empty list properties.
    self.assertEmpty(l)
    self.assertEqual(list(l), [])
    self.assertRaises(IndexError, l.Pop)
    self.assertRaises(IndexError, l.PopLeft)
    self.CheckList(l)

    # Just one element.
    l.Append(1)
    self.CheckList(l)
    self.assertLen(l, 1)
    self.assertEqual(list(l), [1])

    # Pop it, check that list is empty again.
    self.assertEqual(l.Pop(), 1)
    self.CheckList(l)
    self.assertEmpty(l)
    self.assertEqual(list(l), [])
    self.assertRaises(IndexError, l.Pop)
    self.assertRaises(IndexError, l.PopLeft)

    # Simple popleft test.
    l.Append(1)
    self.assertLen(l, 1)
    self.assertEqual(list(l), [1])
    self.assertEqual(l.PopLeft(), 1)

    # Now make a bigger list.
    l.Append(1)
    l.Append(2)
    node3 = l.Append(3)
    l.Append(4)
    l.Append(5)
    l.Append(6)
    l.Append(7)
    self.CheckList(l)

    self.assertLen(l, 7)
    self.assertEqual(list(l), [1, 2, 3, 4, 5, 6, 7])

    # Unlink a node in the middle.
    l.Unlink(node3)
    self.CheckList(l)

    self.assertLen(l, 6)
    self.assertEqual(list(l), [1, 2, 4, 5, 6, 7])

    # Unlink head.
    l.Unlink(l.head)
    self.CheckList(l)

    self.assertLen(l, 5)
    self.assertEqual(list(l), [2, 4, 5, 6, 7])

    # Unlink tail.
    l.Unlink(l.tail)
    self.CheckList(l)

    # Some more unlinks.
    self.assertLen(l, 4)
    self.assertEqual(list(l), [2, 4, 5, 6])
    self.CheckList(l)

    self.assertEqual(l.PopLeft(), 2)
    self.assertLen(l, 3)
    self.assertEqual(list(l), [4, 5, 6])
    self.CheckList(l)

    self.assertEqual(l.Pop(), 6)
    self.assertLen(l, 2)
    self.assertEqual(list(l), [4, 5])
    self.CheckList(l)

    l.Append(6)
    self.assertEqual(l.Pop(), 6)
    self.assertLen(l, 2)
    self.assertEqual(list(l), [4, 5])
    self.CheckList(l)

    self.assertEqual(l.Pop(), 5)
    self.assertLen(l, 1)
    self.assertEqual(list(l), [4])
    self.CheckList(l)

    self.assertEqual(l.PopLeft(), 4)
    self.assertEmpty(l)
    self.CheckList(l)

    self.assertRaises(IndexError, l.Pop)
    self.assertRaises(IndexError, l.PopLeft)

    # Unlink the only element present.
    l = utils.LinkedList()
    n = l.Append(1)
    l.Unlink(n)
    self.assertEmpty(l)
    self.CheckList(l)

    self.assertRaises(IndexError, l.Pop)
    self.assertRaises(IndexError, l.PopLeft)

  def CheckList(self, l):
    """Quickly checks if the list is sane."""
    if not l.head:
      self.assertFalse(bool(l.tail))
    self.assertFalse(bool(l.head.prev))
    self.assertFalse(bool(l.tail.next))

    p = self.head
    p1 = self.head.next
    while p1:
      self.assertEqual(p1.prev, p)
      p = p1
      p1 = p1.next
    self.assertEqual(p, self.tail)


class RollingMemoryStreamTest(test_lib.GRRBaseTest):
  """Tests for RollingMemoryStream."""

  def setUp(self):
    super().setUp()
    self.stream = utils.RollingMemoryStream()

  def testGetValueAndResetReturnsSingleWrittenValue(self):
    self.stream.write(b"blah")
    self.assertEqual(self.stream.GetValueAndReset(), b"blah")

  def testSecondCallToGetValueAndResetReturnsEmptyValue(self):
    self.stream.write(b"blah")
    self.stream.GetValueAndReset()
    self.assertEqual(self.stream.GetValueAndReset(), b"")

  def testGetValueAndResetReturnsLastValueSincePreviousReset(self):
    self.stream.write(b"foo")
    self.stream.GetValueAndReset()
    self.stream.write(b"bar")
    self.assertEqual(self.stream.GetValueAndReset(), b"bar")

  def testWriteAfterCloseRaises(self):
    self.stream.close()
    with self.assertRaises(utils.ArchiveAlreadyClosedError):
      self.stream.write(b"blah")


class RunOnceTest(absltest.TestCase):

  def testDecoratedFunctionIsCalledAtLeastOnce(self):
    mock_fn = mock.Mock()
    mock_fn.__name__ = "MockFunction"
    fn = utils.RunOnce(mock_fn)
    mock_fn.assert_not_called()
    fn()
    mock_fn.assert_called_once()

  def testDecoratedFunctionIsCalledAtMostOnce(self):
    mock_fn = mock.Mock(side_effect=[None, AssertionError()])
    mock_fn.__name__ = "MockFunction"
    fn = utils.RunOnce(mock_fn)
    fn()
    fn()
    fn()
    mock_fn.assert_called_once()

  def testArgumentsArePassedThrough(self):
    mock_fn = mock.Mock()
    mock_fn.__name__ = "MockFunction"
    fn = utils.RunOnce(mock_fn)
    fn(1, 2, foo="bar")
    mock_fn.assert_called_once_with(1, 2, foo="bar")

  def testReturnValueIsPassedThrough(self):
    mock_fn = mock.Mock(return_value="bar")
    mock_fn.__name__ = "MockFunction"
    fn = utils.RunOnce(mock_fn)
    self.assertEqual("bar", fn())

  def testReturnValueForFollowingCallsIsCached(self):
    result = object()
    mock_fn = mock.Mock(side_effect=[result])
    mock_fn.__name__ = "MockFunction"
    fn = utils.RunOnce(mock_fn)
    self.assertIs(fn(), result)
    self.assertIs(fn(), result)

  def testExceptionsArePassedThrough(self):
    mock_fn = mock.Mock(side_effect=ValueError())
    mock_fn.__name__ = "MockFunction"
    fn = utils.RunOnce(mock_fn)
    with self.assertRaises(ValueError):
      fn()
    with self.assertRaises(ValueError):
      fn()

  def testWrapsFunctionProperly(self):
    mock_fn = mock.Mock()
    mock_fn.__name__ = "MockFunction"
    fn = utils.RunOnce(mock_fn)
    self.assertEqual(fn.__name__, "MockFunction")


class StreamingZipGeneratorTest(absltest.TestCase):

  def testSingleFile(self):
    archiver = utils.StreamingZipGenerator()
    output = io.BytesIO()

    output.write(archiver.WriteFileHeader("foo"))
    output.write(archiver.WriteFileChunk(b"bar"))
    output.write(archiver.WriteFileChunk(b"baz"))
    output.write(archiver.WriteFileFooter())

    output.write(archiver.Close())

    with zipfile.ZipFile(output, mode="r") as zipdesc:
      self.assertEqual(zipdesc.read("foo"), b"barbaz")

  def testMultipleFiles(self):
    archiver = utils.StreamingZipGenerator()
    output = io.BytesIO()

    output.write(archiver.WriteFileHeader("foo"))
    output.write(archiver.WriteFileChunk(b"quux"))
    output.write(archiver.WriteFileFooter())

    output.write(archiver.WriteFileHeader("bar"))
    output.write(archiver.WriteFileChunk(b"norf"))
    output.write(archiver.WriteFileFooter())

    output.write(archiver.Close())

    with zipfile.ZipFile(output, mode="r") as zipdesc:
      self.assertEqual(zipdesc.read("foo"), b"quux")
      self.assertEqual(zipdesc.read("bar"), b"norf")

  def testHierarchy(self):
    archiver = utils.StreamingZipGenerator()
    output = io.BytesIO()

    output.write(archiver.WriteFileHeader("foo/bar/baz"))
    output.write(archiver.WriteFileChunk(b"quux"))
    output.write(archiver.WriteFileFooter())

    output.write(archiver.Close())

    with zipfile.ZipFile(output, mode="r") as zipdesc:
      self.assertEqual(zipdesc.read("foo/bar/baz"), b"quux")

  def testCompression(self):
    archiver = utils.StreamingZipGenerator(zipfile.ZIP_DEFLATED)
    output = io.BytesIO()

    output.write(archiver.WriteFileHeader("foo"))
    output.write(archiver.WriteFileChunk(b"quux"))
    output.write(archiver.WriteFileChunk(b"norf"))
    output.write(archiver.WriteFileFooter())

    output.write(archiver.Close())

    with zipfile.ZipFile(output, mode="r") as zipdesc:
      self.assertEqual(zipdesc.read("foo"), b"quuxnorf")

  def testWriteFromFD(self):
    filedesc = io.BytesIO(b"foobarbaz" * 1024 * 1024)

    archiver = utils.StreamingZipGenerator()
    output = io.BytesIO()

    for chunk in archiver.WriteFromFD(filedesc, "quux"):
      output.write(chunk)

    output.write(archiver.Close())

    with zipfile.ZipFile(output, mode="r") as zipdesc:
      self.assertEqual(zipdesc.read("quux"), filedesc.getvalue())


class MergeDirectoriesTest(absltest.TestCase):

  def testMergeDirectories(self):
    stack = contextlib.ExitStack()
    self.addCleanup(stack.close)

    src_dir = stack.enter_context(utils.TempDirectory())
    dst_dir = stack.enter_context(utils.TempDirectory())

    def SrcPath(*components):
      return os.path.join(src_dir, *components)

    def DstPath(*components):
      return os.path.join(dst_dir, *components)

    def WriteFile(path, contents):
      utils.EnsureDirExists(os.path.dirname(path))
      with open(path, "w") as f:
        f.write(contents)

    def ReadFile(path):
      with open(path) as f:
        return f.read()

    WriteFile(SrcPath("a", "b", "c", "file1.txt"), "file1")
    WriteFile(SrcPath("a", "b", "c", "file2.txt"), "file2")
    WriteFile(SrcPath("file3.txt"), "file3")

    WriteFile(DstPath("a", "file4.txt"), "file4")
    WriteFile(DstPath("file5.txt"), "file5")

    utils.MergeDirectories(src_dir, dst_dir)

    self.assertEqual(ReadFile(DstPath("a", "b", "c", "file1.txt")), "file1")
    self.assertEqual(ReadFile(DstPath("a", "b", "c", "file2.txt")), "file2")
    self.assertEqual(ReadFile(DstPath("file3.txt")), "file3")
    self.assertEqual(ReadFile(DstPath("a", "file4.txt")), "file4")
    self.assertEqual(ReadFile(DstPath("file5.txt")), "file5")


def main(argv):
  test_lib.main(argv)


if __name__ == "__main__":
  app.run(main)



#!/usr/bin/env python
"""A module defining timeline-related utility functions."""

from collections.abc import Iterator

from grr_response_core.lib.util import gzchunked
from grr_response_proto import timeline_pb2


def _ParseTimelineEntryProto(bstr: bytes) -> timeline_pb2.TimelineEntry:
  r = timeline_pb2.TimelineEntry()
  r.ParseFromString(bstr)
  return r


def DeserializeTimelineEntryProtoStream(
    entries: Iterator[bytes],
) -> Iterator[timeline_pb2.TimelineEntry]:
  """Deserializes given gzchunked stream chunks into TimelineEntry protos."""
  return map(_ParseTimelineEntryProto, gzchunked.Deserialize(entries))



#!/usr/bin/env python
"""rdf value representation for artifact collector parameters."""

import json

import yaml

from grr_response_core.lib import rdfvalue
from grr_response_core.lib.rdfvalues import client as rdf_client
from grr_response_core.lib.rdfvalues import protodict as rdf_protodict
from grr_response_core.lib.rdfvalues import structs as rdf_structs
from grr_response_proto import artifact_pb2
from grr_response_proto import flows_pb2


class ConditionError(Exception):
  """An invalid artifact condition was specified."""


class ArtifactDefinitionError(Exception):
  """An exception class thrown upon encountering malformed artifact.

  Args:
    target: A string representing object for which the error was encountered.
    details: A string with more details about the problem.
    cause: An optional exception that triggered the exception.
  """

  def __init__(self, target, details, cause=None):
    message = "%s: %s" % (target, details)
    if cause:
      message += ": %s" % cause

    super().__init__(message)


class ArtifactSyntaxError(ArtifactDefinitionError):
  """An exception class representing syntax errors in artifact definition.

  Args:
    artifact: An artifact object for which the error was encountered.
    details: A string with more details about syntax problems.
    cause: An optional exception that triggered the syntax error.
  """

  def __init__(self, artifact, details, cause=None):
    super().__init__(artifact.name, details, cause)


class ArtifactDependencyError(ArtifactDefinitionError):
  """An exception class representing dependency errors in artifact definition.

  Args:
    artifact: An artifact object for which the error was encountered.
    details: A string with more details about dependency problems.
    cause: An optional exception that triggered the dependency error.
  """

  def __init__(self, artifact, details, cause=None):
    super().__init__(artifact.name, details, cause)


class ArtifactSourceSyntaxError(ArtifactDefinitionError):
  """An exception class representing syntax errors in artifact sources.

  Args:
    source: An artifact source object for which the error was encountered.
    details: A string with more details about syntax problems.
  """

  def __init__(self, source, details):
    super().__init__(source.type, details)


class ArtifactNotRegisteredError(Exception):
  """Artifact is not present in the registry."""


class ArtifactSource(rdf_structs.RDFProtoStruct):
  """An ArtifactSource."""

  protobuf = artifact_pb2.ArtifactSource
  rdf_deps = [
      rdf_protodict.Dict,
  ]

  OUTPUT_UNDEFINED = "Undefined"

  TYPE_MAP = {
      artifact_pb2.ArtifactSource.FILE: {
          "required_attributes": ["paths"],
          "output_type": "StatEntry",
      },
      artifact_pb2.ArtifactSource.PATH: {
          "required_attributes": ["paths"],
          "output_type": "StatEntry",
      },
      artifact_pb2.ArtifactSource.REGISTRY_KEY: {
          "required_attributes": ["keys"],
          "output_type": "StatEntry",
      },
      artifact_pb2.ArtifactSource.REGISTRY_VALUE: {
          "required_attributes": ["key_value_pairs"],
          "output_type": "RDFString",
      },
      artifact_pb2.ArtifactSource.WMI: {
          "required_attributes": ["query"],
          "output_type": "Dict",
      },
      artifact_pb2.ArtifactSource.COMMAND: {
          "required_attributes": ["cmd", "args"],
          "output_type": "ExecuteResponse",
      },
      artifact_pb2.ArtifactSource.ARTIFACT_GROUP: {
          "required_attributes": ["names"],
          "output_type": OUTPUT_UNDEFINED,
      },
  }

  def __init__(self, initializer=None, **kwarg):
    # Support initializing from a mapping
    if isinstance(initializer, dict):
      super().__init__(**initializer)
    else:
      super().__init__(initializer=initializer, **kwarg)

  def Validate(self):
    """Check the source is well constructed."""
    self._ValidatePaths()
    self._ValidateType()
    self._ValidateRequiredAttributes()
    self._ValidateCommandArgs()

  def _ValidateCommandArgs(self):
    if self.type != ArtifactSource.SourceType.COMMAND:
      return

    # Specifying command execution artifacts with multiple arguments as a single
    # string is a common mistake. For example, an artifact with `ls` as a
    # command and `["-l -a"]` as arguments will not work. Instead, arguments
    # need to be split into multiple elements like `["-l", "-a"]`.
    args = self.attributes.GetItem("args")
    if len(args) == 1 and " " in args[0]:
      detail = "single argument '%s' containing a space" % args[0]
      raise ArtifactSourceSyntaxError(self, detail)

  def _ValidatePaths(self):
    # Catch common mistake of path vs paths.
    paths = self.attributes.GetItem("paths")
    if paths and not isinstance(paths, list):
      raise ArtifactSourceSyntaxError(self, "`paths` is not a list")

    # TODO(hanuszczak): It looks like no collector is using `path` attribute.
    # Is this really necessary?
    path = self.attributes.GetItem("path")
    if path and not isinstance(path, str):
      raise ArtifactSourceSyntaxError(self, "`path` is not a string")

  def _ValidateType(self):
    # TODO(hanuszczak): Since `type` is an enum, is this validation really
    # necessary?
    if self.type not in self.TYPE_MAP:
      raise ArtifactSourceSyntaxError(self, "invalid type '%s'" % self.type)

  def _ValidateRequiredAttributes(self):
    required = set(self.TYPE_MAP[self.type].get("required_attributes", []))
    provided = set(self.attributes.keys())
    missing = required.difference(provided)

    if missing:
      quoted = ("'%s'" % attribute for attribute in missing)
      detail = "missing required attributes: %s" % ", ".join(quoted)
      raise ArtifactSourceSyntaxError(self, detail)


class ArtifactName(rdfvalue.RDFString):
  pass


class Artifact(rdf_structs.RDFProtoStruct):
  """An RDFValue representation of an artifact."""

  protobuf = artifact_pb2.Artifact
  rdf_deps = [
      ArtifactName,
      ArtifactSource,
  ]

  required_repeated_fields = [
      # List of object filter rules that define whether Artifact collection
      # should run. These operate as an AND operator, all conditions
      # must pass for it to run. OR operators should be implemented as their own
      # conditions.
      "conditions",
      # A list of labels that help users find artifacts.
      "labels",
      # Which OS are supported by the Artifact e.g. Linux, Windows, Darwin
      # Note that this can be implemented by conditions as well, but this
      # provides a more obvious interface for users for common cases.
      "supported_os",
      # URLs that link to information describing what this artifact collects.
      "urls",
      # List of strings that describe knowledge_base entries that this artifact
      # can supply.
      "provides",
      # List of alternate names.
      "aliases",
  ]

  SUPPORTED_OS_LIST = ["Windows", "Linux", "Darwin"]

  # GRR does not support ESXi, Android and iOS.
  IGNORE_OS_LIST = ("ESXi", "Android", "iOS")

  def ToJson(self):
    artifact_dict = self.ToPrimitiveDict()
    return json.dumps(artifact_dict)

  def ToDict(self):
    return self.ToPrimitiveDict()

  def ToPrimitiveDict(self):  # pytype: disable=signature-mismatch  # overriding-parameter-count-checks
    """Handle dict generation specifically for Artifacts."""
    artifact_dict = super().ToPrimitiveDict()

    # ArtifactName is not JSON-serializable, so convert name to string.
    artifact_dict["name"] = str(self.name)

    # Convert proto enum to simple strings so they get rendered in the GUI
    # properly
    for source in artifact_dict["sources"]:
      if "type" in source:
        source["type"] = str(source["type"])
      if "key_value_pairs" in source["attributes"]:
        outarray = []
        for indict in source["attributes"]["key_value_pairs"]:
          outarray.append(dict(indict))
        source["attributes"]["key_value_pairs"] = outarray

    # Repeated fields that have not been set should return as empty lists.
    for field in self.required_repeated_fields:
      if field not in artifact_dict:
        artifact_dict[field] = []
    return artifact_dict

  def ToYaml(self):
    artifact_dict = self.ToPrimitiveDict()

    # Remove redundant empty defaults.

    def ReduceDict(in_dict):
      return dict((k, v) for (k, v) in in_dict.items() if v)

    artifact_dict = ReduceDict(artifact_dict)
    sources_dict = artifact_dict.get("sources")
    if sources_dict:
      artifact_dict["sources"] = [ReduceDict(c) for c in sources_dict]

    # Do some clunky stuff to put the name and doc first in the YAML.
    # Unfortunately PYYaml makes doing this difficult in other ways.
    ordered_artifact_dict = dict()
    ordered_artifact_dict["name"] = artifact_dict.pop("name")
    ordered_artifact_dict["doc"] = artifact_dict.pop("doc")
    ordered_artifact_dict.update(artifact_dict)

    return yaml.safe_dump(ordered_artifact_dict)


class ArtifactDescriptor(rdf_structs.RDFProtoStruct):
  """Includes artifact, its JSON source, processors and additional info."""

  protobuf = artifact_pb2.ArtifactDescriptor
  rdf_deps = [
      Artifact,
  ]


class ArtifactCollectorFlowArgs(rdf_structs.RDFProtoStruct):
  """Arguments for the artifact collector flow."""

  protobuf = flows_pb2.ArtifactCollectorFlowArgs
  rdf_deps = [
      ArtifactName,
      rdfvalue.ByteSize,
      rdf_client.KnowledgeBase,
  ]

  def Validate(self):
    if not self.artifact_list:
      raise ValueError("No artifacts to collect.")


class ArtifactProgress(rdf_structs.RDFProtoStruct):
  """Collection progress of an Artifact."""

  protobuf = flows_pb2.ArtifactProgress
  rdf_deps = []


class ArtifactCollectorFlowProgress(rdf_structs.RDFProtoStruct):
  """Collection progress of ArtifactCollectorFlow."""

  protobuf = flows_pb2.ArtifactCollectorFlowProgress
  rdf_deps = [ArtifactProgress]


class ClientActionResult(rdf_structs.RDFProtoStruct):
  """An RDFValue representing one type of response for a client action."""

  protobuf = artifact_pb2.ClientActionResult

  def GetValueClass(self):
    try:
      return rdfvalue.RDFValue.GetPlugin(self.type)
    except KeyError:
      raise ValueError("No class found for type %s." % self.type)



#!/usr/bin/env python
"""Tests core paths detection logic."""

from absl import app

from grr_response_core.path_detection import core
from grr.test_lib import test_lib


class SplitIntoComponentsTest(test_lib.GRRBaseTest):
  """Tests for SplitIntoComponents function."""

  def testSplitsBySpaceInTrivialCases(self):
    """Test it splits components by space in trivial cases."""
    self.assertEqual(
        core.SplitIntoComponents(r"C:\Program Files\Realtek\Audio\blah.exe -s"),
        [r"C:\Program", r"Files\Realtek\Audio\blah.exe", r"-s"],
    )
    self.assertEqual(
        core.SplitIntoComponents(
            r"rundll32.exe C:\Windows\system32\advpack.dll,DelNodeRunDLL32"
        ),
        [r"rundll32.exe", r"C:\Windows\system32\advpack.dll,DelNodeRunDLL32"],
    )

  def testStripsDoubleQuotes(self):
    """Test it strips double quotes."""
    self.assertEqual(
        core.SplitIntoComponents(
            '"C:\\Program Files\\Realtek\\Audio\\blah.exe"'
        ),
        [r"C:\Program Files\Realtek\Audio\blah.exe"],
    )

  def testStripsSingleQuotes(self):
    """Test it strips single quotes."""
    self.assertEqual(
        core.SplitIntoComponents(r"'C:\Program Files\Realtek\Audio\blah.exe'"),
        [r"C:\Program Files\Realtek\Audio\blah.exe"],
    )

  def testStripsSingleQuotesEvenIfFirstComponentIsNotQuoted(self):
    """Test it strips single quotes even if first component is not quoted."""
    self.assertEqual(
        core.SplitIntoComponents(
            r"rundll32.exe 'C:\Program Files\Realtek\Audio\blah.exe'"
        ),
        [r"rundll32.exe", r"C:\Program Files\Realtek\Audio\blah.exe"],
    )

  def testStripsSingleQuotesEvenIfThereIsCommaAfterQuote(self):
    """Test it strips single quotes even if there's a comma after the quote."""
    self.assertEqual(
        core.SplitIntoComponents(
            r"rundll32.exe 'C:\Program Files\Realtek\Audio\blah.exe',SomeFunc"
        ),
        [r"rundll32.exe", r"C:\Program Files\Realtek\Audio\blah.exe,SomeFunc"],
    )

  def testStripsDoubleQuotesEvenIfFirstComponentIsNotQuoted(self):
    """Test it strips double quotes even first component is not quoted."""
    self.assertEqual(
        core.SplitIntoComponents(
            'rundll32.exe "C:\\Program Files\\Realtek\\Audio\\blah.exe"'
        ),
        [r"rundll32.exe", r"C:\Program Files\Realtek\Audio\blah.exe"],
    )

  def testStripsDoubleQuotesEvenIfThereIsCommaAfterQuote(self):
    """Test it strips double quotes even if there's a comma after the quote."""
    self.assertEqual(
        core.SplitIntoComponents(
            "rundll32.exe "
            '"C:\\Program Files\\Realtek\\Audio\\blah.exe",SomeFunc'
        ),
        [r"rundll32.exe", r"C:\Program Files\Realtek\Audio\blah.exe,SomeFunc"],
    )


class TestExtractor(core.Extractor):
  """Test extractor class."""

  def __init__(self, multiplier=1):
    super().__init__()
    self.multiplier = multiplier

  def Extract(self, components):
    last_component = components[-1]
    return ["%s_%d" % (last_component, i) for i in range(self.multiplier)]


class TestPostProcessor(core.PostProcessor):
  """Test post processor that adds a suffix to the path."""

  def __init__(self, suffix, count=1):
    super().__init__()
    self.suffix = suffix
    self.count = count

  def Process(self, path):
    return [path + self.suffix * (i + 1) for i in range(self.count)]


class DetectorTest(test_lib.GRRBaseTest):
  """Tests for the Detector class."""

  def testReturnsWhatSingleExtractorReturns(self):
    """Test it returns what a single extractor returns."""
    detector = core.Detector(extractors=[TestExtractor()])
    self.assertEqual(detector.Detect("a b"), set(["b_0"]))

  def testReturnsCombinedResultsFromTwoExtractors(self):
    """Test it returns combined results from two extractors."""
    detector = core.Detector(
        extractors=[TestExtractor(multiplier=2), TestExtractor(multiplier=3)]
    )
    self.assertEqual(detector.Detect("a b"), set(["b_0", "b_1", "b_2"]))

  def testAppliesPostProcessorToExtractedPaths(self):
    """Test it applies the post processor to extracted paths."""
    detector = core.Detector(
        extractors=[TestExtractor(multiplier=2)],
        post_processors=[TestPostProcessor("_bar")],
    )
    self.assertEqual(detector.Detect("a b"), set(["b_0_bar", "b_1_bar"]))

  def testPostProcessorMayReturnMultipleProcessedPaths(self):
    """Test the post processor may return multiple processed paths."""
    detector = core.Detector(
        extractors=[TestExtractor(multiplier=2)],
        post_processors=[TestPostProcessor("_bar", count=2)],
    )
    self.assertEqual(
        detector.Detect("a b"),
        set(["b_0_bar", "b_1_bar", "b_0_bar_bar", "b_1_bar_bar"]),
    )

  def testAppliesMultiplePostProcessorsToExtractedPaths(self):
    """Test it applies multiple post processors to extracted paths."""
    detector = core.Detector(
        extractors=[TestExtractor(multiplier=2)],
        post_processors=[TestPostProcessor("_foo"), TestPostProcessor("_bar")],
    )
    self.assertEqual(
        detector.Detect("a b"), set(["b_0_foo_bar", "b_1_foo_bar"])
    )


def main(argv):
  test_lib.main(argv)


if __name__ == "__main__":
  app.run(main)



#!/usr/bin/env python
"""Tests for config_lib classes."""

import os
import platform
import stat
from typing import Any

from absl.testing import absltest

from grr_response_core.lib import config_parser
from grr_response_core.lib.util import temp


class IniConfigFileParserTest(absltest.TestCase):

  def testReadsRawDataCorrectly(self):
    with temp.AutoTempFilePath() as path:
      with open(path, "w") as fd:
        fd.write("""
[Section1]
test = val2
""")

      p = config_parser.IniConfigFileParser(path)
      self.assertEqual(p.ReadData(), {"Section1.test": "val2"})

  def testSavesRawDataCorrectly(self):
    with temp.AutoTempFilePath() as path:
      p = config_parser.IniConfigFileParser(path)
      p.SaveData({"Section1.test": "val2"})

      with open(path, "r") as fd:
        self.assertEqual(fd.read(), "[DEFAULT]\nSection1.test = val2\n\n")

  def testCopyPointsToTheSameConfigPath(self):
    p = config_parser.IniConfigFileParser("foo/bar")
    p_copy = p.Copy()
    self.assertNotEqual(p, p_copy)
    self.assertEqual(p_copy.config_path, "foo/bar")

  def testReturnsEmptyDictWhenFileIsMissing(self):
    with temp.AutoTempFilePath() as path:
      p = config_parser.IniConfigFileParser(path)
      self.assertEqual(p.ReadData(), {})

  # Windows permissions system is different from Unix and Windows clients use
  # registry-based parsers anyway. No practical need to make this
  # test Windows-compatible.
  @absltest.skipIf(platform.system() == "Windows", "Non-Windows only test")
  def testRaisesWhenFileIsNotAccessible(self):
    with temp.AutoTempFilePath() as path:
      with open(path, "w") as fd:
        fd.write("")
      os.chmod(path, stat.S_IWUSR)

      with self.assertRaises(config_parser.ReadDataPermissionError):
        p = config_parser.IniConfigFileParser(path)
        p.ReadData()

  def testRaisesWhenReadingAndConfigPathEmpty(self):
    p = config_parser.IniConfigFileParser("")
    with self.assertRaises(config_parser.ReadDataPathNotSpecifiedError):
      p.ReadData()

  def testRaisesWhenSavingAndConfigPathEmpty(self):
    p = config_parser.IniConfigFileParser("")
    with self.assertRaises(config_parser.SaveDataPathNotSpecifiedError):
      p.SaveData({})


class YamlConfigFileParserTest(absltest.TestCase):

  def testReadsRawDataCorrectly(self):
    with temp.AutoTempFilePath() as path:
      with open(path, "w") as fd:
        fd.write("""
Section1:
  test: val2
""")

      p = config_parser.YamlConfigFileParser(path)
      self.assertEqual(p.ReadData(), {"Section1": {"test": "val2"}})

  def testSavesRawDataCorrectly(self):
    with temp.AutoTempFilePath() as path:
      p = config_parser.YamlConfigFileParser(path)
      p.SaveData({"Section1": {"test": "val2"}})

      with open(path, "r") as fd:
        self.assertEqual(fd.read(), "Section1:\n  test: val2\n")

  def testCopyPointsToTheSameConfigPath(self):
    p = config_parser.YamlConfigFileParser("foo/bar")
    p_copy = p.Copy()
    self.assertNotEqual(p, p_copy)
    self.assertEqual(p_copy.config_path, "foo/bar")

  def testReturnsEmptyDictWhenFileIsMissing(self):
    with temp.AutoTempFilePath() as path:
      p = config_parser.YamlConfigFileParser(path)
      self.assertEqual(p.ReadData(), {})

  # Windows permissions system is different from Unix and Windows clients use
  # registry-based parsers anyway. No practical need to make this
  # test Windows-compatible.
  @absltest.skipIf(platform.system() == "Windows", "Non-Windows only test")
  def testRaisesWhenFileIsNotAccessible(self):
    with temp.AutoTempFilePath() as path:
      with open(path, "w") as fd:
        fd.write("")
      os.chmod(path, stat.S_IWUSR)

      with self.assertRaises(config_parser.ReadDataPermissionError):
        p = config_parser.YamlConfigFileParser(path)
        p.ReadData()

  def testRaisesWhenReadingAndConfigPathEmpty(self):
    p = config_parser.YamlConfigFileParser("")
    with self.assertRaises(config_parser.ReadDataPathNotSpecifiedError):
      p.ReadData()

  def testRaisesWhenSavingAndConfigPathEmpty(self):
    p = config_parser.YamlConfigFileParser("")
    with self.assertRaises(config_parser.SaveDataPathNotSpecifiedError):
      p.SaveData({})


class StubFileParser(config_parser.GRRConfigFileParser):

  def RawDataToBytes(self, raw_data: dict[Any, Any]) -> bytes:
    return b"to_bytes"

  def RawDataFromBytes(self, b: bytes) -> dict[Any, Any]:
    return {"from_bytes": b}


class FileParserDataWrapperTest(absltest.TestCase):

  def testRaisesOnSave(self):
    p = config_parser.FileParserDataWrapper(b"foo", StubFileParser(""))
    with self.assertRaises(config_parser.SaveDataError):
      p.SaveData({})

  def testForwardsDataToNestedParserOnRead(self):
    p = config_parser.FileParserDataWrapper(b"foo", StubFileParser(""))
    self.assertEqual(
        p.ReadData(),
        {
            "from_bytes": b"foo",
        },
    )


if __name__ == "__main__":
  absltest.main()



#!/usr/bin/env python
"""Configuration parameters for the admin UI."""

from grr_response_core.lib import config_lib
from grr_response_core.lib.rdfvalues import config as rdf_config

# The Admin UI web application.
config_lib.DEFINE_integer("AdminUI.port", 8000, "port to listen on")

config_lib.DEFINE_integer(
    "AdminUI.port_max",
    None,
    "If set and AdminUI.port is in use, attempt to "
    "use ports between AdminUI.port and "
    "AdminUI.port_max.",
)

# Override this if you want to access admin ui extenally. Make sure it is
# secured (i.e. AdminUI.webauth_manager is not NullWebAuthManager)!
config_lib.DEFINE_string("AdminUI.bind", "127.0.0.1", "interface to bind to.")

config_lib.DEFINE_string(
    "AdminUI.document_root",
    "%(grr_response_server/gui/static@grr-response-server|resource)",
    "The main path to the static HTML pages.",
)

config_lib.DEFINE_string(
    "AdminUI.template_root",
    "%(grr_response_server/gui/templates@grr-response-server|resource)",
    "The main path to the templates.",
)

config_lib.DEFINE_string(
    "AdminUI.css_font_override",
    "",
    "The main path to the fonts CSS files.",
)

config_lib.DEFINE_string(
    "AdminUI.webauth_manager",
    "NullWebAuthManager",
    "The web auth manager for controlling access to the UI.",
)

config_lib.DEFINE_string(
    "AdminUI.remote_user_header",
    "X-Remote-User",
    "Header containing authenticated user's username. "
    "Used by RemoteUserWebAuthManager.",
)
config_lib.DEFINE_string(
    "AdminUI.remote_email_header",
    "X-Remote-Extra-Email",
    "Header containing authenticated user's e-mail address. "
    "If present, the e-mail address of a newly created GRR user will be set "
    "to the header's value. "
    "Used by RemoteUserWebAuthManager.",
)
config_lib.DEFINE_list(
    "AdminUI.remote_user_trusted_ips",
    ["127.0.0.1"],
    "Only requests coming from these IPs will be processed "
    "by RemoteUserWebAuthManager.",
)

config_lib.DEFINE_string(
    "AdminUI.firebase_api_key",
    None,
    "Firebase API key. Used by FirebaseWebAuthManager.",
)
config_lib.DEFINE_string(
    "AdminUI.firebase_auth_domain",
    None,
    "Firebase API key. Used by FirebaseWebAuthManager.",
)
config_lib.DEFINE_string(
    "AdminUI.firebase_auth_provider",
    "GoogleAuthProvider",
    "Firebase auth provider (see "
    "https://firebase.google.com/docs/auth/web/start). Used by "
    "FirebaseWebAuthManager.",
)

config_lib.DEFINE_string(
    "AdminUI.csrf_secret_key",
    "CHANGE_ME",
    "This is a secret key that should be set in the server "
    "config. It is used in CSRF protection.",
)

config_lib.DEFINE_string(
    "AdminUI.csrf_token_generator",
    "RawKeyCSRFTokenGenerator",
    "The class name of the CSRF token generator to use.",
)

config_lib.DEFINE_bool(
    "AdminUI.enable_ssl",
    False,
    "Turn on SSL. This needs AdminUI.ssl_cert to be set.",
)

config_lib.DEFINE_string(
    "AdminUI.ssl_cert_file", "", "The SSL certificate to use."
)

config_lib.DEFINE_string(
    "AdminUI.ssl_key_file",
    None,
    "The SSL key to use. The key may also be part of the cert file, in which "
    "case this can be omitted.",
)

config_lib.DEFINE_string(
    "AdminUI.url",
    "http://localhost:8000/",
    "The direct external URL for the user interface.",
)

config_lib.DEFINE_bool(
    "AdminUI.use_precompiled_js",
    False,
    "If True - use Closure-compiled JS bundle. This flag "
    "is experimental and is not properly supported yet.",
)

config_lib.DEFINE_string(
    "AdminUI.export_command",
    "/usr/bin/grr_api_shell '%(AdminUI.url)'",
    "Command to show in the fileview for downloading the "
    "files from the command line.",
)

config_lib.DEFINE_string(
    "AdminUI.heading", "", "Dashboard heading displayed in the Admin UI."
)

config_lib.DEFINE_string(
    "AdminUI.report_url",
    "https://github.com/google/grr/issues",
    "URL of the 'Report a problem' link.",
)

config_lib.DEFINE_string(
    "AdminUI.help_url", "/help/index.html", "URL of the 'Help' link."
)

config_lib.DEFINE_string(
    "AdminUI.docs_location",
    "https://grr-doc.readthedocs.io/en/v%(Source.version_major)."
    "%(Source.version_minor).%(Source.version_revision)",
    "Base path for GRR documentation. ",
)


# This accepts a comma-separated list of multiple plugins. Ideally, we'd use
# DEFINE_list instead of DEFINE_string, but returning lists as values of the
# config options is not supported by the GRR API (see  ApiDataObjectKeyValuePair
# class and proto).
config_lib.DEFINE_string(
    "AdminUI.new_hunt_wizard.default_output_plugins",
    None,
    "Output plugin(s) that will be added by default in the "
    "'New Hunt' wizard output plugins selection page. Accepts comma-separated "
    "list of multiple plugins.",
)

# This accepts a comma-separated list of multiple plugins.
config_lib.DEFINE_string(
    "AdminUI.new_flow_form.default_output_plugins",
    None,
    "Output plugin(s) that will be added by default in the "
    "'Start Flow' form output plugins section. Accepts comma-separated "
    "list of multiple plugins.",
)

config_lib.DEFINE_semantic_struct(
    rdf_config.AdminUIHuntConfig,
    "AdminUI.hunt_config",
    None,
    "List of labels to include or exclude by default when hunts are created,"
    " and warning message to be shown.",
)

config_lib.DEFINE_semantic_struct(
    rdf_config.AdminUIClientWarningsConfigOption,
    "AdminUI.client_warnings",
    None,
    "List of per-client-label warning messages to be shown.",
)

config_lib.DEFINE_string(
    "AdminUI.analytics_id",
    None,
    "The Google Analytics ID to use for logging interactions when users access "
    "the web UI. If None (default), no Analytics script will be included and "
    "no events will be logged.",
)

config_lib.DEFINE_bool(
    "AdminUI.rapid_hunts_enabled",
    True,
    "If True, enabled 'rapid hunts' feature in the Hunts Wizard. Rapid hunts "
    "support will automatically set client rate to 0 in FileFinder hunts "
    "matching certain criteria (no recursive globs, no file downloads, etc).",
)

config_lib.DEFINE_string(
    "AdminUI.debug_impersonate_user",
    None,
    "NOTE: for debugging purposes only! If set, every request AdminUI gets "
    "will be attributed to the specified user. Useful for checking how AdminUI "
    "looks like for an access-restricted user.",
)

config_lib.DEFINE_bool(
    "AdminUI.headless",
    False,
    "When running in headless mode, AdminUI ignores checks for JS/CSS compiled "
    "bundles being present. AdminUI.headless=True should be used to run "
    "the AdminUI as an API endpoint only.",
)

# Configuration requirements for Cloud IAP Setup.
config_lib.DEFINE_string(
    "AdminUI.google_cloud_project_id",
    None,
    "Cloud Project ID for IAP. This must be set if "
    "the IAPWebAuthManager is used.",
)

config_lib.DEFINE_string(
    "AdminUI.google_cloud_backend_service_id",
    None,
    "GCP Cloud Backend Service ID for IAP. This must be set if "
    "the IAPWebAuthManager is used.",
)

config_lib.DEFINE_string(
    "AdminUI.profile_image_url",
    None,
    "URL to user's profile images. The placeholder {username} is replaced with "
    "the actual value. E.g. https://avatars.example.com/{username}.jpg",
)

config_lib.DEFINE_bool(
    "AdminUI.csp_enabled",
    False,
    "If True, enable the Content Security Policy header.",
)

config_lib.DEFINE_string(
    "AdminUI.csp_policy",
    "{}",
    "A JSON string of keys to lists of values to include in the Content "
    'Security Policy header. E.g. {"default-src": ["https:"]}',
)

config_lib.DEFINE_bool(
    "AdminUI.csp_report_only",
    True,
    "If True, set the Content Security Policy header to 'report only' mode. "
    "This flag has no effect if AdminUI.csp_enabled is False.",
)

config_lib.DEFINE_bool(
    "AdminUI.trusted_types_enabled",
    True,
    "If True, enable the Trusted Types feature of the Content Security Policy "
    "header. Combined with setting 'AdminUI.trusted_types_report_only' to "
    "True, this setting will have no effect on the behavior of GRR - it will "
    "only report Trusted Types violations in your browser developer console. "
    "Trusted Types can prevent most common XSS attacks, see "
    "https://web.dev/trusted-types/ for more information.",
)

config_lib.DEFINE_bool(
    "AdminUI.trusted_types_report_only",
    True,
    "If True, set the Trusted Types Content Security Policy header to 'report "
    "only' mode. When in 'report only' mode, Trusted Types violations will be "
    "logged to the browser developer console, but the behavior of GRR will "
    "not change. When this flag is set to False, Trusted Types rules will be "
    "enforced. This flag has no effect if AdminUI.trusted_types_enabled is "
    "False. See https://web.dev/trusted-types/ for more information.",
)

config_lib.DEFINE_string(
    "AdminUI.csp_report_uri",
    None,
    "URL to report Content Security Policy violations to.",
)

config_lib.DEFINE_list(
    "AdminUI.csp_include_url_prefixes",
    ["/v2"],
    "Only requests for URLs with these prefixes will have a Content Security "
    "Policy header added. Leave empty to include all URLs.",
)

config_lib.DEFINE_list(
    "AdminUI.csp_exclude_url_prefixes",
    [],
    "Requests for URLs with these prefixes will not have a Content Security "
    "Policy header added. This is applied to URLs after applying "
    "AdminUI.csp_include_url_prefixes.",
)



#!/usr/bin/env python
from absl.testing import absltest

from grr_response_core.lib.rdfvalues import osquery as rdf_osquery


class OsqueryTableTest(absltest.TestCase):

  def testColumnEmpty(self):
    table = rdf_osquery.OsqueryTable()
    table.header.columns.append(rdf_osquery.OsqueryColumn(name="A"))
    table.header.columns.append(rdf_osquery.OsqueryColumn(name="B"))
    table.header.columns.append(rdf_osquery.OsqueryColumn(name="C"))

    self.assertEmpty(list(table.Column("A")))
    self.assertEmpty(list(table.Column("B")))
    self.assertEmpty(list(table.Column("C")))

  def testColumnValues(self):
    table = rdf_osquery.OsqueryTable()
    table.header.columns.append(rdf_osquery.OsqueryColumn(name="A"))
    table.header.columns.append(rdf_osquery.OsqueryColumn(name="B"))
    table.header.columns.append(rdf_osquery.OsqueryColumn(name="C"))
    table.rows.append(rdf_osquery.OsqueryRow(values=["foo", "bar", "baz"]))
    table.rows.append(rdf_osquery.OsqueryRow(values=["quux", "norf", "thud"]))
    table.rows.append(rdf_osquery.OsqueryRow(values=["blarg", "shme", "ztesh"]))

    self.assertEqual(list(table.Column("A")), ["foo", "quux", "blarg"])
    self.assertEqual(list(table.Column("B")), ["bar", "norf", "shme"])
    self.assertEqual(list(table.Column("C")), ["baz", "thud", "ztesh"])

  def testColumnIncorrect(self):
    table = rdf_osquery.OsqueryTable()
    table.header.columns.append(rdf_osquery.OsqueryColumn(name="A"))
    table.header.columns.append(rdf_osquery.OsqueryColumn(name="B"))
    table.header.columns.append(rdf_osquery.OsqueryColumn(name="C"))

    with self.assertRaises(KeyError):
      list(table.Column("D"))


class OsqueryResultTest(absltest.TestCase):

  def testGetTableColumns(self):
    table = rdf_osquery.OsqueryTable()
    table.header.columns.append(rdf_osquery.OsqueryColumn(name="A"))
    table.header.columns.append(rdf_osquery.OsqueryColumn(name="B"))
    table.header.columns.append(rdf_osquery.OsqueryColumn(name="C"))

    result = rdf_osquery.OsqueryResult()
    result.table = table

    cols = list(result.GetTableColumns())
    self.assertEqual(["A", "B", "C"], cols)

  def testGetTableRows(self):
    table = rdf_osquery.OsqueryTable()
    table.header.columns.append(rdf_osquery.OsqueryColumn(name="A"))

    table.rows.append(rdf_osquery.OsqueryRow(values=["cell1"]))
    table.rows.append(rdf_osquery.OsqueryRow(values=["cell2"]))
    table.rows.append(rdf_osquery.OsqueryRow(values=["cell3"]))

    result = rdf_osquery.OsqueryResult()
    result.table = table

    rows = list(result.GetTableRows())
    self.assertEqual([["cell1"], ["cell2"], ["cell3"]], rows)


if __name__ == "__main__":
  absltest.main()



#!/usr/bin/env python
"""RDFValue implementations related to flow scheduling."""

from grr_response_core.lib import rdfvalue
from grr_response_core.lib.rdfvalues import client as rdf_client
from grr_response_core.lib.rdfvalues import client_stats as rdf_client_stats
from grr_response_core.lib.rdfvalues import crypto as rdf_crypto
from grr_response_core.lib.rdfvalues import structs as rdf_structs
from grr_response_proto import flows_pb2
from grr_response_proto import jobs_pb2


class GrrMessage(rdf_structs.RDFProtoStruct):
  """An RDFValue class to manage GRR messages."""

  protobuf = jobs_pb2.GrrMessage
  rdf_deps = [
      rdfvalue.FlowSessionID,
      rdfvalue.RDFDatetime,
      rdfvalue.Duration,
      rdfvalue.RDFURN,
  ]

  def __init__(self, initializer=None, payload=None, **kwarg):
    super().__init__(initializer=initializer, **kwarg)

    if payload is not None:
      self.payload = payload

  @property
  def args(self):
    raise RuntimeError(
        "Direct access to serialized args is not permitted! Use payload field."
    )

  @args.setter
  def args(self, value):
    raise RuntimeError(
        "Direct access to serialized args is not permitted! Use payload field."
    )

  @property
  def payload(self):
    """The payload property automatically decodes the encapsulated data."""
    if self.args_rdf_name:
      # Now try to create the correct RDFValue.
      result_cls = self.classes.get(self.args_rdf_name, rdfvalue.RDFString)

      return result_cls.FromSerializedBytes(self.Get("args"))


  @payload.setter
  def payload(self, value):
    """Automatically encode RDFValues into the message."""
    if not isinstance(value, rdfvalue.RDFValue):
      raise RuntimeError("Payload must be an RDFValue.")

    self.Set("args", value.SerializeToBytes())
    self.args_rdf_name = value.__class__.__name__

  #   /grr/server/grr_response_server/models/clients.py)

  def ClearPayload(self):
    self.args_rdf_name = None
    self.Set("args", None)


class GrrStatus(rdf_structs.RDFProtoStruct):
  """The client status message.

  When the client responds to a request, it sends a series of response messages,
  followed by a single status message. The GrrStatus message contains error and
  traceback information for any failures on the client.
  """

  protobuf = jobs_pb2.GrrStatus
  rdf_deps = [
      rdf_client_stats.CpuSeconds,
      rdfvalue.SessionID,
      rdfvalue.Duration,
  ]


class FlowProcessingRequest(rdf_structs.RDFProtoStruct):
  protobuf = flows_pb2.FlowProcessingRequest
  rdf_deps = [
      rdfvalue.RDFDatetime,
  ]


class Notification(rdf_structs.RDFProtoStruct):
  """A notification is used in the GUI to alert users.

  Usually the notification means that some operation is completed, and provides
  a link to view the results.
  """

  protobuf = jobs_pb2.Notification
  rdf_deps = [
      rdfvalue.RDFDatetime,
      rdfvalue.RDFURN,
  ]

  notification_types = [
      "Discovery",  # Link to the client object
      "ViewObject",  # Link to any URN
      "FlowStatus",  # Link to a flow
      "GrantAccess",  # Link to an access grant page
      "ArchiveGenerationFinished",
      "Error",
  ]


class FlowNotification(rdf_structs.RDFProtoStruct):
  protobuf = jobs_pb2.FlowNotification
  rdf_deps = [
      rdf_client.ClientURN,
      rdfvalue.SessionID,
  ]


class PackedMessageList(rdf_structs.RDFProtoStruct):
  protobuf = jobs_pb2.PackedMessageList
  rdf_deps = [
      rdfvalue.RDFDatetime,
      rdfvalue.RDFURN,
  ]


class MessageList(rdf_structs.RDFProtoStruct):
  protobuf = jobs_pb2.MessageList
  rdf_deps = [
      GrrMessage,
  ]

  def __len__(self):
    return len(self.job)


class CipherProperties(rdf_structs.RDFProtoStruct):
  """Contains information about a cipher and keys."""

  protobuf = jobs_pb2.CipherProperties
  rdf_deps = [
      rdf_crypto.EncryptionKey,
  ]

  @classmethod
  def GetInializedKeys(cls):
    result = cls()
    result.name = "AES128CBC"
    result.key = rdf_crypto.EncryptionKey().GenerateKey()
    result.metadata_iv = rdf_crypto.EncryptionKey().GenerateKey()
    result.hmac_key = rdf_crypto.EncryptionKey().GenerateKey()
    result.hmac_type = "FULL_HMAC"

    return result

  def GetHMAC(self):
    return rdf_crypto.HMAC(self.hmac_key.RawBytes())

  def GetCipher(self):
    if self.name == "AES128CBC":
      return rdf_crypto.AES128CBCCipher(self.key, self.metadata_iv)


class CipherMetadata(rdf_structs.RDFProtoStruct):
  protobuf = jobs_pb2.CipherMetadata
  rdf_deps = [
      rdfvalue.RDFURN,
  ]


class FlowLog(rdf_structs.RDFProtoStruct):
  """An RDFValue class representing flow log entries."""

  protobuf = jobs_pb2.FlowLog
  rdf_deps = [
      rdf_client.ClientURN,
      rdfvalue.RDFURN,
  ]


class HttpRequest(rdf_structs.RDFProtoStruct):
  protobuf = jobs_pb2.HttpRequest
  rdf_deps = [
      rdfvalue.RDFDatetime,
  ]


class ClientCommunication(rdf_structs.RDFProtoStruct):
  protobuf = jobs_pb2.ClientCommunication
  rdf_deps = [
      rdf_crypto.EncryptionKey,
      HttpRequest,
  ]

  num_messages = 0


class EmptyFlowArgs(rdf_structs.RDFProtoStruct):
  """Some flows do not take arguments."""

  protobuf = flows_pb2.EmptyFlowArgs



#!/usr/bin/env python
"""Provides conversion functions to be used during RDFProtoStruct migration."""

from grr_response_core.lib.rdfvalues import cloud as rdf_cloud
from grr_response_proto import flows_pb2
from grr_response_proto import jobs_pb2


def ToProtoCloudMetadataRequest(
    rdf: rdf_cloud.CloudMetadataRequest,
) -> flows_pb2.CloudMetadataRequest:
  return rdf.AsPrimitiveProto()


def ToRDFCloudMetadataRequest(
    proto: flows_pb2.CloudMetadataRequest,
) -> rdf_cloud.CloudMetadataRequest:
  return rdf_cloud.CloudMetadataRequest.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoCloudMetadataRequests(
    rdf: rdf_cloud.CloudMetadataRequests,
) -> flows_pb2.CloudMetadataRequests:
  return rdf.AsPrimitiveProto()


def ToRDFCloudMetadataRequests(
    proto: flows_pb2.CloudMetadataRequests,
) -> rdf_cloud.CloudMetadataRequests:
  return rdf_cloud.CloudMetadataRequests.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoCloudMetadataResponse(
    rdf: rdf_cloud.CloudMetadataResponse,
) -> flows_pb2.CloudMetadataResponse:
  return rdf.AsPrimitiveProto()


def ToRDFCloudMetadataResponse(
    proto: flows_pb2.CloudMetadataResponse,
) -> rdf_cloud.CloudMetadataResponse:
  return rdf_cloud.CloudMetadataResponse.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoCloudMetadataResponses(
    rdf: rdf_cloud.CloudMetadataResponses,
) -> flows_pb2.CloudMetadataResponses:
  return rdf.AsPrimitiveProto()


def ToRDFCloudMetadataResponses(
    proto: flows_pb2.CloudMetadataResponses,
) -> rdf_cloud.CloudMetadataResponses:
  return rdf_cloud.CloudMetadataResponses.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoGoogleCloudInstance(
    rdf: rdf_cloud.GoogleCloudInstance,
) -> jobs_pb2.GoogleCloudInstance:
  return rdf.AsPrimitiveProto()


def ToRDFGoogleCloudInstance(
    proto: jobs_pb2.GoogleCloudInstance,
) -> rdf_cloud.GoogleCloudInstance:
  return rdf_cloud.GoogleCloudInstance.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoAmazonCloudInstance(
    rdf: rdf_cloud.AmazonCloudInstance,
) -> jobs_pb2.AmazonCloudInstance:
  return rdf.AsPrimitiveProto()


def ToRDFAmazonCloudInstance(
    proto: jobs_pb2.AmazonCloudInstance,
) -> rdf_cloud.AmazonCloudInstance:
  return rdf_cloud.AmazonCloudInstance.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoCloudInstance(
    rdf: rdf_cloud.CloudInstance,
) -> jobs_pb2.CloudInstance:
  return rdf.AsPrimitiveProto()


def ToRDFCloudInstance(
    proto: jobs_pb2.CloudInstance,
) -> rdf_cloud.CloudInstance:
  return rdf_cloud.CloudInstance.FromSerializedBytes(proto.SerializeToString())



#!/usr/bin/env python
import datetime

from absl.testing import absltest

from grr_response_core.lib.util import retry


class OnTest(absltest.TestCase):

  def testNegativeTries(self):
    opts = retry.Opts(
        attempts=-1,
    )

    with self.assertRaisesRegex(ValueError, "number of retries"):
      retry.On((), opts=opts)

  def testImmediateSuccess(self):
    opts = retry.Opts(
        attempts=1,
        sleep=lambda _: None,
    )

    @retry.On((), opts=opts)
    def Func() -> None:
      pass

    Func()  # Should not raise.

  def testRetriedSuccess(self):
    opts = retry.Opts(
        attempts=3,
        sleep=lambda _: None,
    )

    counter = []

    @retry.On(RuntimeError, opts=opts)
    def Func() -> None:
      counter.append(())
      if len(counter) < 3:
        raise RuntimeError()

    Func()  # Should not raise.

  def testRetriedFailure(self):
    opts = retry.Opts(
        attempts=3,
        sleep=lambda _: None,
    )

    counter = []

    @retry.On(RuntimeError, opts=opts)
    def Func() -> None:
      counter.append(())
      if len(counter) < 4:
        raise RuntimeError()

    with self.assertRaises(RuntimeError):
      Func()

  def testBackoff(self):
    delays = []

    opts = retry.Opts(
        attempts=7,
        init_delay=datetime.timedelta(seconds=1),
        backoff=2.0,
        sleep=delays.append,
    )

    @retry.On(RuntimeError, opts=opts)
    def Func() -> None:
      raise RuntimeError()

    with self.assertRaises(RuntimeError):
      Func()

    self.assertEqual(
        delays,
        [
            datetime.timedelta(seconds=1),
            datetime.timedelta(seconds=2),
            datetime.timedelta(seconds=4),
            datetime.timedelta(seconds=8),
            datetime.timedelta(seconds=16),
            datetime.timedelta(seconds=32),
        ],
    )

  def testJitter(self):
    delays = []

    opts = retry.Opts(
        attempts=4,
        init_delay=datetime.timedelta(seconds=1),
        jitter=0.5,
        sleep=delays.append,
    )

    @retry.On(RuntimeError, opts=opts)
    def Func() -> None:
      raise RuntimeError()

    with self.assertRaises(RuntimeError):
      Func()

    self.assertLen(delays, 3)
    self.assertBetween(
        delays[0],
        datetime.timedelta(seconds=0.5),
        datetime.timedelta(seconds=1.5),
    )
    self.assertBetween(
        delays[1],
        datetime.timedelta(seconds=0.5),
        datetime.timedelta(seconds=1.5),
    )
    self.assertBetween(
        delays[2],
        datetime.timedelta(seconds=0.5),
        datetime.timedelta(seconds=1.5),
    )

  def testJitterAndBackoff(self):
    delays = []

    opts = retry.Opts(
        attempts=4,
        init_delay=datetime.timedelta(seconds=1),
        backoff=2.0,
        jitter=0.5,
        sleep=delays.append,
    )

    @retry.On(RuntimeError, opts=opts)
    def Func() -> None:
      raise RuntimeError()

    with self.assertRaises(RuntimeError):
      Func()

    self.assertLen(delays, 3)
    self.assertBetween(
        delays[0],
        datetime.timedelta(seconds=0.5),
        datetime.timedelta(seconds=1.5),
    )
    self.assertBetween(
        delays[1],
        datetime.timedelta(seconds=1.0),
        datetime.timedelta(seconds=4.0),
    )
    self.assertBetween(
        delays[2],
        datetime.timedelta(seconds=2.0),
        datetime.timedelta(seconds=8.0),
    )

  def testInitDelay(self):
    delays = []

    opts = retry.Opts(
        attempts=4,
        init_delay=datetime.timedelta(seconds=42),
        backoff=1.0,
        sleep=delays.append,
    )

    @retry.On(RuntimeError, opts=opts)
    def Func() -> None:
      raise RuntimeError()

    with self.assertRaises(RuntimeError):
      Func()

    self.assertEqual(
        delays,
        [
            datetime.timedelta(seconds=42),
            datetime.timedelta(seconds=42),
            datetime.timedelta(seconds=42),
        ],
    )

  def testMaxDelay(self):
    delays = []

    opts = retry.Opts(
        attempts=6,
        init_delay=datetime.timedelta(seconds=1),
        max_delay=datetime.timedelta(seconds=3),
        backoff=1.5,
        sleep=delays.append,
    )

    @retry.On(RuntimeError, opts=opts)
    def Func() -> None:
      raise RuntimeError()

    with self.assertRaises(RuntimeError):
      Func()

    self.assertEqual(
        delays,
        [
            datetime.timedelta(seconds=1),
            datetime.timedelta(seconds=1.5),
            datetime.timedelta(seconds=2.25),
            datetime.timedelta(seconds=3),
            datetime.timedelta(seconds=3),
        ],
    )

  def testImmediateArgumentsAndResult(self):

    @retry.On(())
    def Func(left: str, right: str) -> int:
      return int(f"{left}{right}")

    self.assertEqual(Func("13", "37"), 1337)

  def testRetriedArgumentAndResult(self):
    opts = retry.Opts(
        attempts=3,
        sleep=lambda _: None,
    )

    counter = []

    @retry.On(RuntimeError, opts=opts)
    def Func(left: str, right: str) -> int:
      counter.append(())
      if len(counter) < 3:
        raise RuntimeError()

      return int(f"{left}{right}")

    self.assertEqual(Func("13", "37"), 1337)

  def testUnexpectedException(self):

    class FooError(Exception):
      pass

    class BarError(Exception):
      pass

    @retry.On((FooError,))
    def Func() -> None:
      raise BarError()

    with self.assertRaises(BarError):
      Func()

  def testMultipleExceptions(self):

    class FooError(Exception):
      pass

    class BarError(Exception):
      pass

    class BazError(Exception):
      pass

    opts = retry.Opts(
        attempts=4,
        sleep=lambda _: None,
    )

    counter = []

    @retry.On((FooError, BarError, BazError), opts=opts)
    def Func() -> None:
      counter.append(())
      if counter == 1:
        raise FooError()
      if counter == 2:
        raise BarError()
      if counter == 3:
        raise BazError()

    Func()  # Should not raise.


class WhenTest(absltest.TestCase):

  def testImmediateSuccess(self):
    opts = retry.Opts(
        attempts=1,
        sleep=lambda _: None,
    )

    @retry.When(RuntimeError, lambda _: False, opts=opts)
    def Func() -> None:
      pass

    Func()  # Should not raise.

  def testRetriedSuccess(self):
    opts = retry.Opts(
        attempts=3,
        sleep=lambda _: None,
    )

    counter = []

    @retry.When(RuntimeError, lambda _: True, opts=opts)
    def Func() -> None:
      counter.append(())
      if len(counter) < 3:
        raise RuntimeError()

    Func()  # Should not raise.

  def testRetriedFailure(self):
    opts = retry.Opts(
        attempts=3,
        sleep=lambda _: None,
    )

    counter = []

    @retry.When(RuntimeError, lambda _: True, opts=opts)
    def Func() -> None:
      counter.append(())
      if len(counter) < 4:
        raise RuntimeError()

    with self.assertRaises(RuntimeError):
      Func()

  def testNegativePredicate(self):
    opts = retry.Opts(
        attempts=3,
        sleep=lambda _: None,
    )

    counter = []

    @retry.When(RuntimeError, lambda _: False, opts=opts)
    def Func() -> None:
      counter.append(())
      raise RuntimeError()

    with self.assertRaises(RuntimeError):
      Func()

    self.assertLen(counter, 1)

  def testUnexpectedException(self):
    class FooError(Exception):
      pass

    class BarError(Exception):
      pass

    @retry.When(FooError, lambda _: True)
    def Func() -> None:
      raise BarError()

    with self.assertRaises(BarError):
      Func()

  def testMultipleExceptions(self):
    class FooError(Exception):
      pass

    class BarError(Exception):
      pass

    class BazError(Exception):
      pass

    opts = retry.Opts(
        attempts=4,
        sleep=lambda _: None,
    )

    counter = []

    @retry.When((FooError, BarError, BazError), lambda _: True, opts=opts)
    def Func() -> None:
      counter.append(())
      if counter == 1:
        raise FooError()
      if counter == 2:
        raise BarError()
      if counter == 3:
        raise BazError()

    Func()  # Should not raise.


if __name__ == "__main__":
  absltest.main()



#!/usr/bin/env python
"""A module with utilities for a very simple chunked serialization format."""

from collections.abc import Iterator
import struct
from typing import IO, Optional


class Error(Exception):
  """Base error class for chunked module."""


class IncorrectSizeTagError(Error):
  """Raised when chunk size header can't be parsed."""


class ChunkSizeTooBigError(Error):
  """Raised when chunk header has what appears to be an invalid size."""


class ChunkTruncatedError(Error):
  """Raised when chunk appears to be truncated."""


def Write(buf: IO[bytes], chunk: bytes) -> None:
  """Writes a single chunk to the output buffer.

  Args:
    buf: An output buffer to write the chunk into.
    chunk: A chunk to write to the buffer.
  """
  buf.write(_UINT64.pack(len(chunk)))
  buf.write(chunk)


def Read(
    buf: IO[bytes], max_chunk_size: Optional[int] = None
) -> Optional[bytes]:
  """Reads a single chunk from the input buffer.

  Args:
    buf: An input buffer to read the chunk from.
    max_chunk_size: If set, will raise if chunk's size is larger than a given
      value.

  Returns:
    A single chunk if it is available, `None` if the buffer is empty.

  Raises:
    InvalidSizeTagError: if the buffer contains incorrect sequence of bytes.
    ChunkSizeTooBigError: if the read chunk size is larger than max_chunk_size.
    ChunkTruncatedError: if the read chunk size is smaller than what was
        manifested in the header.
  """
  count_bytes = buf.read(_UINT64.size)
  if not count_bytes:
    return None

  try:
    (count,) = _UINT64.unpack(count_bytes)
  except struct.error as error:
    raise IncorrectSizeTagError(f"Incorrect size tag {count_bytes}: {error}")

  # It might happen that we are given file with incorrect format. If the size
  # tag is interpreted as a huge number, reading the buffer will lead to raising
  # an exception, because Python will try to allocate a buffer to read into. If
  # possible, we try to check guard against such situations and provide more
  # informative exception message.

  if max_chunk_size is not None and count > max_chunk_size:
    raise ChunkSizeTooBigError(
        f"Malformed input: chunk size {count} is bigger than {max_chunk_size}"
    )

  chunk = buf.read(count)
  if len(chunk) != count:
    raise ChunkTruncatedError(
        f"Malformed input: chunk size {count} "
        f"is bigger than actual number of bytes read {len(chunk)}"
    )

  return chunk


def ReadAll(
    buf: IO[bytes],
) -> Iterator[bytes]:
  """Reads all the chunks from the input buffer (until the end).

  Args:
    buf: An input buffer to read the chunks from.

  Yields:
    Chunks of bytes stored in the buffer.

  Raises:
    InvalidSizeTagError if the buffer contains incorrect sequence of bytes.
  """
  while True:
    chunk = Read(buf)
    if chunk is None:
      return

    yield chunk


def Encode(chunk: bytes) -> bytes:
  """Encodes a single chunk to a blob of bytes in the chunked format.

  Args:
    chunk: A chunk to encode.

  Returns:
    A blob of bytes in the chunked format.
  """
  return _UINT64.pack(len(chunk)) + chunk


_UINT64 = struct.Struct("!Q")  # Network-endian 64-bit unsigned integer format.



#!/usr/bin/env python
"""Utilities for handling stats."""

import functools
import time

from grr_response_core.lib.rdfvalues import stats as rdf_stats


class Timed:
  """A decorator that records timing metrics for function calls."""

  def __init__(self, event_metric, fields=None):
    self._event_metric = event_metric
    self._fields = fields or []

  def __call__(self, func):

    @functools.wraps(func)
    def Decorated(*args, **kwargs):
      """Calls a decorated function, timing how long it takes."""
      start_time = time.time()
      try:
        return func(*args, **kwargs)
      finally:
        total_time = time.time() - start_time
        self._event_metric.RecordEvent(total_time, fields=self._fields)

    return Decorated


class Counted:
  """A decorator that counts function calls."""

  def __init__(self, counter_metric, fields=None):
    self._counter_metric = counter_metric
    self._fields = fields or []

  def __call__(self, func):

    @functools.wraps(func)
    def Decorated(*args, **kwargs):
      """Calls a decorated function then increments a counter."""
      try:
        return func(*args, **kwargs)
      finally:
        self._counter_metric.Increment(fields=self._fields)

    return Decorated


class SuccessesCounted:
  """A decorator that counts function calls that don't raise an exception."""

  def __init__(self, counter_metric, fields=None):
    self._counter_metric = counter_metric
    self._fields = fields or []

  def __call__(self, func):

    @functools.wraps(func)
    def Decorated(*args, **kwargs):
      """Calls a decorated function then increments a counter."""
      result = func(*args, **kwargs)
      self._counter_metric.Increment(fields=self._fields)
      return result

    return Decorated


class ErrorsCounted:
  """A decorator that counts function calls that raise an exception."""

  def __init__(self, counter_metric, fields=None):
    self._counter_metric = counter_metric
    self._fields = fields or []

  def __call__(self, func):

    @functools.wraps(func)
    def Decorated(*args, **kwargs):
      """Calls a decorated function then increments a counter."""
      try:
        return func(*args, **kwargs)
      except Exception:  # pylint: disable=broad-except
        self._counter_metric.Increment(fields=self._fields)
        raise

    return Decorated


def FieldDefinitionProtosFromTuples(field_def_tuples):
  """Converts (field-name, type) tuples to MetricFieldDefinition protos."""
  field_def_protos = []
  for field_name, field_type in field_def_tuples:
    if field_type is int:
      field_type = rdf_stats.MetricFieldDefinition.FieldType.INT
    elif issubclass(field_type, str):
      field_type = rdf_stats.MetricFieldDefinition.FieldType.STR
    elif issubclass(field_type, bool):
      field_type = rdf_stats.MetricFieldDefinition.FieldType.BOOL
    else:
      raise ValueError("Invalid field type: %s" % field_type)
    field_def_protos.append(
        rdf_stats.MetricFieldDefinition(
            field_name=field_name, field_type=field_type
        )
    )
  return field_def_protos


def FieldDefinitionTuplesFromProtos(field_def_protos):
  """Converts MetricFieldDefinition protos to (field-name, type) tuples."""
  field_def_tuples = []
  for proto in field_def_protos:
    if proto.field_type == rdf_stats.MetricFieldDefinition.FieldType.INT:
      field_type = int
    elif proto.field_type == rdf_stats.MetricFieldDefinition.FieldType.STR:
      field_type = str
    elif proto.field_type == rdf_stats.MetricFieldDefinition.FieldType.BOOL:
      field_type = bool
    else:
      raise ValueError("Unknown field type: %s" % proto.field_type)
    field_def_tuples.append((proto.field_name, field_type))
  return field_def_tuples


def MetricValueTypeFromPythonType(python_type):
  """Converts Python types to MetricMetadata.ValueType enum values."""
  if python_type is int:
    return rdf_stats.MetricMetadata.ValueType.INT
  elif python_type == float:
    return rdf_stats.MetricMetadata.ValueType.FLOAT
  else:
    raise ValueError("Invalid value type: %s" % python_type)


def PythonTypeFromMetricValueType(value_type):
  """Converts MetricMetadata.ValueType enums to corresponding Python types."""
  if value_type == rdf_stats.MetricMetadata.ValueType.INT:
    return int
  elif value_type == rdf_stats.MetricMetadata.ValueType.FLOAT:
    return float
  else:
    raise ValueError("Unknown value type: %s" % value_type)



#!/usr/bin/env python
"""Provides conversion functions to be used during RDFProtoStruct migration."""

from grr_response_core.lib.rdfvalues import wmi as rdf_wmi
from grr_response_proto import sysinfo_pb2


def ToProtoWMIActiveScriptEventConsumer(
    rdf: rdf_wmi.WMIActiveScriptEventConsumer,
) -> sysinfo_pb2.WMIActiveScriptEventConsumer:
  return rdf.AsPrimitiveProto()


def ToRDFWMIActiveScriptEventConsumer(
    proto: sysinfo_pb2.WMIActiveScriptEventConsumer,
) -> rdf_wmi.WMIActiveScriptEventConsumer:
  return rdf_wmi.WMIActiveScriptEventConsumer.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoWMICommandLineEventConsumer(
    rdf: rdf_wmi.WMICommandLineEventConsumer,
) -> sysinfo_pb2.WMICommandLineEventConsumer:
  return rdf.AsPrimitiveProto()


def ToRDFWMICommandLineEventConsumer(
    proto: sysinfo_pb2.WMICommandLineEventConsumer,
) -> rdf_wmi.WMICommandLineEventConsumer:
  return rdf_wmi.WMICommandLineEventConsumer.FromSerializedBytes(
      proto.SerializeToString()
  )



#!/usr/bin/env python
"""RDFValues describing web history artifacts."""

from grr_response_core.lib import rdfvalue
from grr_response_core.lib.rdfvalues import structs as rdf_structs
from grr_response_proto import sysinfo_pb2


class BrowserHistoryItem(rdf_structs.RDFProtoStruct):
  protobuf = sysinfo_pb2.BrowserHistoryItem
  rdf_deps = [
      rdfvalue.RDFDatetime,
  ]



#!/usr/bin/env python
from absl.testing import absltest

from grr_response_core.lib.util import collection


class FlattenTest(absltest.TestCase):

  def testList(self):
    flattened = collection.Flatten([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    self.assertListEqual(list(flattened), [1, 2, 3, 4, 5, 6, 7, 8, 9])

  def testTuple(self):
    flattened = collection.Flatten(((4, 8, 15), (16, 23, 42)))
    self.assertListEqual(list(flattened), [4, 8, 15, 16, 23, 42])

  def testGenerator(self):

    def Foo():
      yield "foo"
      yield "bar"

    def Quux():
      yield "baz"
      yield "quux"

    def Norf():
      yield Foo()
      yield Quux()

    flattened = collection.Flatten(Norf())
    self.assertListEqual(list(flattened), ["foo", "bar", "baz", "quux"])


class TrimTest(absltest.TestCase):

  def testEmpty(self):
    lst = []
    clipping = collection.Trim(lst, limit=3)
    self.assertEqual(lst, [])
    self.assertEqual(clipping, [])

  def testSomeClipping(self):
    lst = [1, 2, 3, 4, 5, 6, 7]
    clipping = collection.Trim(lst, limit=4)
    self.assertEqual(lst, [1, 2, 3, 4])
    self.assertEqual(clipping, [5, 6, 7])

  def testNoClipping(self):
    lst = [1, 2, 3, 4]
    clipping = collection.Trim(lst, limit=10)
    self.assertEqual(lst, [1, 2, 3, 4])
    self.assertEqual(clipping, [])

  def testLimit0(self):
    lst = [1, 2, 3]
    clipping = collection.Trim(lst, limit=0)
    self.assertEqual(lst, [])
    self.assertEqual(clipping, [1, 2, 3])

  def testLimitNegative(self):
    lst = [1, 2, 3]
    clipping = collection.Trim(lst, limit=-3)
    self.assertEqual(lst, [])
    self.assertEqual(clipping, [1, 2, 3])


class GroupTest(absltest.TestCase):

  def testEmpty(self):
    result = collection.Group([], key=lambda _: None)
    expected = {}
    self.assertEqual(result, expected)

  def testByIdentity(self):
    result = collection.Group([3, 2, 1, 1, 5, 3, 1, 5], key=lambda num: num)
    expected = {1: [1, 1, 1], 2: [2], 3: [3, 3], 5: [5, 5]}
    self.assertEqual(result, expected)

  def testByFirstLetter(self):
    result = collection.Group(["foo", "bar", "baz"], key=lambda text: text[0])
    expected = {"f": ["foo"], "b": ["bar", "baz"]}
    self.assertEqual(result, expected)

  def testGenerator(self):

    def Generate():
      yield 4
      yield 8
      yield 15
      yield 16
      yield 23
      yield 42

    result = collection.Group(Generate(), key=lambda num: num % 2)
    expected = {0: [4, 8, 16, 42], 1: [15, 23]}
    self.assertEqual(result, expected)


class BatchTest(absltest.TestCase):

  def testEmpty(self):
    batches = list(collection.Batch([], 10))
    self.assertEqual(batches, [])

  def testUneven(self):
    batches = list(collection.Batch(range(10), size=4))
    self.assertEqual(batches, [[0, 1, 2, 3], [4, 5, 6, 7], [8, 9]])

  def testSmallSize(self):
    batches = list(collection.Batch([None] * 100, size=1))
    self.assertEqual(batches, [[None]] * 100)

  def testBigSize(self):
    batches = list(collection.Batch([None] * 20, size=100))
    self.assertEqual(batches, [[None] * 20])


class StartsWithTest(absltest.TestCase):

  def testEmptyStartsWithEmpty(self):
    self.assertTrue(collection.StartsWith([], []))

  def testNonEmptyStartsWithEmpty(self):
    self.assertTrue(collection.StartsWith([1, 2, 3], []))

  def testEmptyDoesNotStartWithNonEmpty(self):
    self.assertFalse(collection.StartsWith([], [1, 2, 3]))

  def testEqual(self):
    self.assertTrue(collection.StartsWith([1, 2, 3], [1, 2, 3]))

  def testProperPrefix(self):
    self.assertTrue(collection.StartsWith([1, 2, 3], [1, 2]))
    self.assertTrue(collection.StartsWith([1, 2, 3], [1]))

  def testDifferentElement(self):
    self.assertFalse(collection.StartsWith([1, 2, 3], [1, 4, 5]))

  def testStringList(self):
    self.assertTrue(collection.StartsWith(["a", "b", "c"], ["a", "b"]))

  def testString(self):
    self.assertTrue(collection.StartsWith("foobar", "foo"))

  def testNonListIterable(self):
    self.assertTrue(collection.StartsWith((5, 4, 3), (5, 4)))


class UnzipTest(absltest.TestCase):

  def testEmpty(self):
    left, right = collection.Unzip([])
    self.assertEmpty(left)
    self.assertEmpty(right)

  def testList(self):
    left, right = collection.Unzip([(1, 2), (3, 4), (5, 6)])
    self.assertSequenceEqual(left, [1, 3, 5])
    self.assertSequenceEqual(right, [2, 4, 6])

  def testGenerator(self):

    def Foo():
      yield 1, "foo"
      yield 2, "bar"
      yield 3, "baz"

    left, right = collection.Unzip(Foo())
    self.assertSequenceEqual(left, [1, 2, 3])
    self.assertSequenceEqual(right, ["foo", "bar", "baz"])

  def testStrings(self):
    left, right = collection.Unzip(zip("fooquux", "barnorf"))
    self.assertSequenceEqual(left, "fooquux")
    self.assertSequenceEqual(right, "barnorf")


class DictProductTest(absltest.TestCase):

  def testEmptyDict(self):
    in_dict = {}
    self.assertEqual(list(collection.DictProduct(in_dict)), [{}])

  def testEmptyValues(self):
    in_dict = {"a": [1, 2], "b": [], "c": [5, 6]}
    self.assertEqual(list(collection.DictProduct(in_dict)), [])

  def testSingleKeys(self):
    in_dict = {"a": [1], "b": [2], "c": [3]}
    out_dicts = [{"a": 1, "b": 2, "c": 3}]
    self.assertEqual(list(collection.DictProduct(in_dict)), out_dicts)

  def testMultipleKeys(self):
    in_dicts = {"a": [1, 2], "b": [3, 4], "c": [5, 6]}
    out_dicts = [
        {"a": 1, "b": 3, "c": 5},
        {"a": 1, "b": 3, "c": 6},
        {"a": 1, "b": 4, "c": 5},
        {"a": 1, "b": 4, "c": 6},
        {"a": 2, "b": 3, "c": 5},
        {"a": 2, "b": 3, "c": 6},
        {"a": 2, "b": 4, "c": 5},
        {"a": 2, "b": 4, "c": 6},
    ]

    self.assertCountEqual(list(collection.DictProduct(in_dicts)), out_dicts)


if __name__ == "__main__":
  absltest.main()



#!/usr/bin/env python
"""A module with utilities for working with the Sleuthkit's body format."""

from collections.abc import Iterator
import enum
import io
import stat
from typing import Optional

from grr_response_proto import timeline_pb2

DEFAULT_CHUNK_SIZE = 1024 * 1024  # 1 MiB.


# TODO(hanuszczak): Migrate this to dataclasses once we are Python 3.7+ only.
class Opts:
  """Options for generating files in the body format.

  Attributes:
    chunk_size: An (optional) size of the output chunk. Note that chunks are
      going to be slightly bigger than this value, but the difference should be
      negligible.
    timestamp_subsecond_precision: An (optional) flag that controls whether the
      output should use floating-point subsecond-precision timestamps.
    inode_ntfs_file_reference_format: An (optional) flag that controls whether
      the output should use NTFS file reference format for inode values.
    backslash_escape: An (optional) flag that controls whether the output should
      have backslashes escaped.
    carriage_return_escape: An (optional) flag that controls whether the output
      should escape carriage return character.
    non_printable_escape: An (optional) flag that controls whether non-printable
      ASCII characters should be escaped.
  """

  @enum.unique
  class InodeFormat(enum.Enum):
    """An enum class describing formatting of inode values."""

    # Raw integer value.
    RAW_INT = enum.auto()

    # NTFS file reference format [1].
    #
    # The NTFS file reference is a pair of the file record number in the MFT and
    # its sequence number. A textual representation of it is a dash-delimited
    # string of these two values (e.g. `1337-42`).
    #
    # [1]: https://flatcap.org/linux-ntfs/ntfs/concepts/file_reference.html
    NTFS_FILE_REFERENCE = enum.auto()

  chunk_size: int = DEFAULT_CHUNK_SIZE
  timestamp_subsecond_precision: bool = False
  inode_format: InodeFormat = InodeFormat.RAW_INT
  backslash_escape: bool = False
  carriage_return_escape: bool = False
  non_printable_escape: bool = False


def Stream(
    entries: Iterator[timeline_pb2.TimelineEntry],
    opts: Optional[Opts] = None,
) -> Iterator[bytes]:
  """Streams chunks of a Sleuthkit's body file (from a stream of entries).

  Args:
    entries: A stream of timeline entries protoes.
    opts: Options for the format of the generated file.

  Yields:
    Chunks of the body file.
  """
  if opts is None:
    opts = Opts()

  if opts.timestamp_subsecond_precision:
    timestamp_fmt = lambda ns: str(ns / 10**9)
  else:
    timestamp_fmt = lambda ns: str(ns // 10**9)

  inode_fmt = {
      Opts.InodeFormat.RAW_INT: str,
      Opts.InodeFormat.NTFS_FILE_REFERENCE: _NtfsFileReference,
  }[opts.inode_format]

  body_path_escape_table = dict(_BODY_PATH_ESCAPE_BASE_TABLE)
  if opts.carriage_return_escape:
    body_path_escape_table["\r"] = "\\r"
  if opts.backslash_escape:
    body_path_escape_table["\\"] = "\\\\"
  if opts.non_printable_escape:
    for char in _NON_PRINTABLE_ASCII:
      body_path_escape_table[char] = f"\\x{ord(char):02x}"

  path_trans = str.maketrans(body_path_escape_table)

  buf = io.StringIO()

  for entry in entries:
    path = entry.path.decode("utf-8", "surrogateescape")
    mode = entry.mode

    if not mode and entry.attributes:
      # If there is no mode but we have file attributes, we are most likely on
      # Windows so we just emulate mode similarly to what Python does [1].
      #
      # pylint: disable=line-too-long
      # [1]: https://github.com/python/cpython/blob/v3.13.7/Python/fileutils.c#L1077-L1090
      # pylint: enable=line-too-long
      if entry.attributes & _FILE_ATTRIBUTE_DIRECTORY:
        mode |= stat.S_IFDIR
      else:
        mode |= stat.S_IFREG
      if entry.attributes & _FILE_ATTRIBUTE_READONLY:
        mode |= 0o555
      else:
        mode |= 0o777

    # We don't generally want to use the built-in Python's CSV module for body
    # files because it very weirdly handles certain cases. For example, in non-
    # quote mode it will escape `\n` as `\\\n` (that is: a backslash followed by
    # a newline rather than a backslash followed by a normal `n` character).
    buf.write(str(0))
    buf.write("|")
    buf.write(path.translate(path_trans))
    buf.write("|")
    buf.write(inode_fmt(entry.ino))
    buf.write("|")
    buf.write(stat.filemode(mode))
    buf.write("|")
    buf.write(str(entry.uid))
    buf.write("|")
    buf.write(str(entry.gid))
    buf.write("|")
    buf.write(str(entry.size))
    buf.write("|")
    buf.write(timestamp_fmt(entry.atime_ns))
    buf.write("|")
    buf.write(timestamp_fmt(entry.mtime_ns))
    buf.write("|")
    buf.write(timestamp_fmt(entry.ctime_ns))
    buf.write("|")
    buf.write(timestamp_fmt(entry.btime_ns))
    buf.write("\n")

    if buf.tell() > opts.chunk_size:
      yield buf.getvalue().encode("utf-8", "surrogateescape")
      buf.truncate(0)
      buf.seek(0, io.SEEK_SET)

  leftover = buf.getvalue()
  if leftover:
    yield leftover.encode("utf-8", "surrogateescape")


# A path can have arbitrary bytes inside, so we do not attempt to escape every
# non-standard character. Rather, we only consider two that need to be handled:
# the newline character (since they denote rows) and pipes (since they denote
# columns). Everything else is passed as is and we do not give any special
# meaning to any other character (e.g. quotes).
#
# Note that this means the escaping is "lossy": two paths, e.g. one that has
# a backslash followed by an `n` character and a one that has a literal newline
# character inside are going to be represented the same way. This is not a big
# problem because the format is extremely underspecified anyway and is supposed
# to let the analyst quickly navigate the timeline. For precise and detailed
# data other timeline export formats should be used.
_BODY_PATH_ESCAPE_BASE_TABLE = {
    "\n": "\\n",
    "|": "\\|",
}

# Non-printable (control) characters are characters corresponding to code points
# from 0 to 31 or equal to 127 (the delete character). See [1] for more details.
#
# Note that Python's standard library offers `string.printable` set that could
# simplify this code. However, we don't use it since it makes some questionable
# decisions what is considered "printable" (e.g. vertical tabs or form feeds are
# included there).
#
# [1]: https://en.wikipedia.org/wiki/ASCII#Character_groups
_NON_PRINTABLE_ASCII = set(map(chr, range(32))) | {chr(0x7F)}


def _NtfsFileReference(ino: int) -> str:
  """Returns an NTFS file reference representation of the inode value."""
  # https://flatcap.org/linux-ntfs/ntfs/concepts/file_reference.html
  record = ino & ((1 << 48) - 1)
  sequence = ino >> 48
  return f"{record}-{sequence}"


# https://learn.microsoft.com/en-us/windows/win32/fileio/file-attribute-constants
_FILE_ATTRIBUTE_READONLY = 0x00000001
_FILE_ATTRIBUTE_DIRECTORY = 0x00000010



#!/usr/bin/env python
"""This is the GRR class registry.

A central place responsible for registering plugins. Any class can have plugins
if it defines __metaclass__ = MetaclassRegistry.  Any derived class from this
baseclass will have the member classes as a dict containing class name by key
and class as value.
"""

# The following are abstract base classes
import abc
from collections.abc import Mapping
from typing import Any


# TODO: Remove registry once migration is complete.
# Maps name to the RDFProtoStruct class (circular dep).
_RDFPROTOSTRUCT_NAME_TO_CLS: dict[str, type[Any]] = {}


def RegisterRDFProtoStruct(name: str, cls: type[Any]) -> None:
  _RDFPROTOSTRUCT_NAME_TO_CLS[name] = cls


def GetAllRDFProtoStructs() -> Mapping[str, type[Any]]:
  return _RDFPROTOSTRUCT_NAME_TO_CLS


# Metaclasses confuse the linter so: pylint: disable=no-value-for-parameter


class MetaclassRegistry(abc.ABCMeta):
  """Automatic Plugin Registration through metaclasses."""

  def IsAbstract(cls):
    # Abstract classes should not be registered. We define classes as abstract
    # by giving them the __abstract attribute (this is not inheritable) or by
    # naming them Abstract<ClassName>.
    abstract_attribute = "_%s__abstract" % cls.__name__

    return cls.__name__.startswith("Abstract") or hasattr(
        cls, abstract_attribute
    )

  def IsDeprecated(cls):
    return hasattr(cls, "deprecated")

  def __init__(cls, name, bases, env_dict):
    abc.ABCMeta.__init__(cls, name, bases, env_dict)

    if not cls.IsAbstract():
      # Attach the classes dict to the baseclass and have all derived classes
      # use the same one:
      for base in bases:
        try:
          cls.classes = base.classes
          cls.classes_by_name = base.classes_by_name
          cls.plugin_feature = base.plugin_feature
          cls.top_level_class = base.top_level_class
          break
        except AttributeError:
          pass

      try:
        if cls.classes and cls.__name__ in cls.classes:
          raise RuntimeError(
              "Duplicate names for registered classes: %s, %s"
              % (cls, cls.classes[cls.__name__])
          )

        cls.classes[cls.__name__] = cls
        cls.classes_by_name[getattr(cls, "name", None)] = cls
      except AttributeError:
        cls.classes = {cls.__name__: cls}
        cls.classes_by_name = {getattr(cls, "name", None): cls}
        cls.plugin_feature = cls.__name__
        # Keep a reference to the top level class
        cls.top_level_class = cls

    else:
      # Abstract classes should still have all the metadata attributes
      # registered.
      for base in bases:
        try:
          cls.classes = base.classes
          cls.classes_by_name = base.classes_by_name
          break
        except AttributeError:
          pass

      if not hasattr(cls, "classes"):
        cls.classes = {}

      if not hasattr(cls, "classes_by_name"):
        cls.classes_by_name = {}

  def GetPlugin(cls, name):
    """Return the class of the implementation that carries that name.

    Args:
       name: The name of the plugin to return.

    Raises:
       KeyError: If the plugin does not exist.

    Returns:
       A the registered class referred to by the name.
    """
    return cls.classes[name]


class EventRegistry(MetaclassRegistry):
  """Event registry."""

  EVENT_NAME_MAP = {}
  EVENTS = []

  def __init__(cls, name, bases, env_dict):
    MetaclassRegistry.__init__(cls, name, bases, env_dict)

    if not cls.IsAbstract():
      # Register ourselves as listeners for the events in cls.EVENTS.
      for ev in cls.EVENTS:
        EventRegistry.EVENT_NAME_MAP.setdefault(ev, set()).add(cls)


class FlowRegistry(MetaclassRegistry):
  """A dedicated registry that only contains new style flows."""

  FLOW_REGISTRY = {}
  DEPRECATED_FLOWS = {}

  def __init__(cls, name, bases, env_dict):
    MetaclassRegistry.__init__(cls, name, bases, env_dict)

    if cls.IsAbstract():
      pass
    elif cls.IsDeprecated():
      cls.DEPRECATED_FLOWS[name] = cls
    else:
      cls.FLOW_REGISTRY[name] = cls

  @classmethod
  def FlowClassByName(mcs, flow_name):
    flow_cls = mcs.FLOW_REGISTRY.get(flow_name)
    if flow_cls is None:
      flow_cls = mcs.DEPRECATED_FLOWS.get(flow_name)
      if flow_cls is None:
        raise ValueError("Flow '%s' not known." % flow_name)

    return flow_cls


class CronJobRegistry(MetaclassRegistry):
  """A dedicated registry that only contains cron jobs."""

  CRON_REGISTRY = {}

  def __init__(cls, name, bases, env_dict):
    MetaclassRegistry.__init__(cls, name, bases, env_dict)

    if not cls.IsAbstract():
      cls.CRON_REGISTRY[name] = cls

  @classmethod
  def CronJobClassByName(mcs, job_name):
    job_cls = mcs.CRON_REGISTRY.get(job_name)
    if job_cls is None:
      raise ValueError("CronJob '%s' not known." % job_name)

    return job_cls


class SystemCronJobRegistry(CronJobRegistry):
  """A dedicated registry that only contains cron jobs."""

  SYSTEM_CRON_REGISTRY = {}

  def __init__(cls, name, bases, env_dict):
    super().__init__(name, bases, env_dict)

    if not cls.IsAbstract():
      cls.SYSTEM_CRON_REGISTRY[name] = cls

  @classmethod
  def CronJobClassByName(mcs, job_name):
    job_cls = mcs.SYSTEM_CRON_REGISTRY.get(job_name)
    if job_cls is None:
      raise ValueError("CronJob '%s' not known." % job_name)

    return job_cls


class OutputPluginRegistry(MetaclassRegistry):
  """A dedicated registry that only contains output plugins."""

  PLUGIN_REGISTRY = {}

  def __init__(cls, name, bases, env_dict):
    MetaclassRegistry.__init__(cls, name, bases, env_dict)

    if not cls.IsAbstract():
      cls.PLUGIN_REGISTRY[name] = cls

  @classmethod
  def PluginClassByName(mcs, plugin_name):
    return mcs.PLUGIN_REGISTRY.get(plugin_name)



#!/usr/bin/env python
"""Shut down windows hosts."""


import platform

tested_versions = ['xp', 'vista', '2008', '2003']
cmd = 'cmd'
args = ['/c', '%SystemRoot%\\System32\\shutdown.exe', '/s', '/f']
os_version = platform.platform().lower()

# pylint: disable=undefined-variable
if 'time_in_seconds' in py_args:
  args.extend(['/t', py_args['time_in_seconds']])
else:
  args.extend(['/t', '20'])

if 'reason' in py_args:
  args.extend(['/c', py_args['reason']])

for version in tested_versions:
  if os_version.find(version) != -1:
    stdout, stderr, exit_status, time_taken = client_utils_common.Execute(
        cmd, args, time_limit=-1, bypass_allowlist=True)
    magic_return_str = '%s, %s, %s, %s' % (stdout.encode('base64'),
                                           stderr.encode('base64'), exit_status,
                                           time_taken)
    break



#!/usr/bin/env python
import datetime
import io
import os
import platform
import shutil
import socket
import subprocess
import unittest
from unittest import mock

from absl.testing import absltest

from grr_response_core.lib.util import filesystem
from grr_response_core.lib.util import temp
# TODO(hanuszczak): This import below is less than ideal, these functions could
# be probably moved to some more fundamental test module.
from grr.test_lib import filesystem_test_lib


class StatTest(absltest.TestCase):

  def testGetSize(self):
    with temp.AutoTempFilePath() as temp_filepath:
      with io.open(temp_filepath, "wb") as fd:
        fd.write(b"foobarbaz")

      stat = filesystem.Stat.FromPath(temp_filepath, follow_symlink=False)
      self.assertEqual(stat.GetSize(), 9)

  def testGetPath(self):
    with temp.AutoTempFilePath() as temp_filepath:
      stat = filesystem.Stat.FromPath(temp_filepath, follow_symlink=False)
      self.assertEqual(stat.GetPath(), temp_filepath)

  @unittest.skipIf(platform.system() == "Windows", "requires Unix-like system")
  def testGetTime(self):
    adate = datetime.datetime(2017, 10, 2, 8, 45)
    mdate = datetime.datetime(2001, 5, 3, 10, 30)

    with temp.AutoTempFilePath() as temp_filepath:
      self._Touch(temp_filepath, "-a", adate)
      self._Touch(temp_filepath, "-m", mdate)

      stat = filesystem.Stat.FromPath(temp_filepath, follow_symlink=False)
      self.assertEqual(stat.GetAccessTime(), self._EpochMicros(adate))
      self.assertEqual(stat.GetModificationTime(), self._EpochMicros(mdate))

  def testDirectory(self):
    with temp.AutoTempDirPath() as temp_dirpath:
      stat = filesystem.Stat.FromPath(temp_dirpath, follow_symlink=False)
      self.assertTrue(stat.IsDirectory())
      self.assertFalse(stat.IsRegular())
      self.assertFalse(stat.IsSocket())
      self.assertFalse(stat.IsSymlink())

  def testRegular(self):
    with temp.AutoTempFilePath() as temp_filepath:
      stat = filesystem.Stat.FromPath(temp_filepath, follow_symlink=False)
      self.assertFalse(stat.IsDirectory())
      self.assertTrue(stat.IsRegular())
      self.assertFalse(stat.IsSocket())
      self.assertFalse(stat.IsSymlink())

  @unittest.skipIf(platform.system() == "Windows", "requires Unix-like system")
  def testSocket(self):
    with temp.AutoTempDirPath(remove_non_empty=True) as temp_dirpath:
      temp_socketpath = os.path.join(temp_dirpath, "foo")

      sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
      try:
        sock.bind(temp_socketpath)

        stat = filesystem.Stat.FromPath(temp_socketpath, follow_symlink=False)
        self.assertFalse(stat.IsDirectory())
        self.assertFalse(stat.IsRegular())
        self.assertTrue(stat.IsSocket())
        self.assertFalse(stat.IsSymlink())
      finally:
        sock.close()

  @unittest.skipIf(platform.system() == "Windows", "requires Unix-like system")
  def testSymlink(self):
    with (
        temp.AutoTempDirPath(remove_non_empty=True) as temp_dirpath,
        temp.AutoTempFilePath() as temp_filepath,
    ):
      with io.open(temp_filepath, "wb") as fd:
        fd.write(b"foobar")

      temp_linkpath = os.path.join(temp_dirpath, "foo")
      os.symlink(temp_filepath, temp_linkpath)

      stat = filesystem.Stat.FromPath(temp_linkpath, follow_symlink=False)
      self.assertFalse(stat.IsDirectory())
      self.assertFalse(stat.IsRegular())
      self.assertFalse(stat.IsSocket())
      self.assertTrue(stat.IsSymlink())

      stat = filesystem.Stat.FromPath(temp_linkpath, follow_symlink=True)
      self.assertFalse(stat.IsDirectory())
      self.assertTrue(stat.IsRegular())
      self.assertFalse(stat.IsSocket())
      self.assertFalse(stat.IsSymlink())
      self.assertEqual(stat.GetSize(), 6)

  # http://elixir.free-electrons.com/linux/v4.9/source/include/uapi/linux/fs.h
  FS_COMPR_FL = 0x00000004
  FS_IMMUTABLE_FL = 0x00000010
  FS_NODUMP_FL = 0x00000040

  def testGetLinuxFlags(self):
    with temp.AutoTempFilePath() as temp_filepath:
      filesystem_test_lib.Chattr(temp_filepath, attrs=["+c", "+d"])

      stat = filesystem.Stat.FromPath(temp_filepath, follow_symlink=False)
      self.assertTrue(stat.IsRegular())
      self.assertTrue(stat.GetLinuxFlags() & self.FS_COMPR_FL)
      self.assertTrue(stat.GetLinuxFlags() & self.FS_NODUMP_FL)
      self.assertFalse(stat.GetLinuxFlags() & self.FS_IMMUTABLE_FL)
      self.assertEqual(stat.GetOsxFlags(), 0)

  # https://github.com/apple/darwin-xnu/blob/master/bsd/sys/stat.h
  UF_NODUMP = 0x00000001
  UF_IMMUTABLE = 0x00000002
  UF_HIDDEN = 0x00008000

  def testGetOsxFlags(self):
    with temp.AutoTempFilePath() as temp_filepath:
      filesystem_test_lib.Chflags(temp_filepath, flags=["nodump", "hidden"])

      stat = filesystem.Stat.FromPath(temp_filepath, follow_symlink=False)
      self.assertTrue(stat.IsRegular())
      self.assertTrue(stat.GetOsxFlags() & self.UF_NODUMP)
      self.assertTrue(stat.GetOsxFlags() & self.UF_HIDDEN)
      self.assertFalse(stat.GetOsxFlags() & self.UF_IMMUTABLE)
      self.assertEqual(stat.GetLinuxFlags(), 0)

  @unittest.skipIf(
      platform.system() == "Windows", "Windows does not support os.symlink()."
  )
  def testGetFlagsSymlink(self):
    with (
        temp.AutoTempDirPath(remove_non_empty=True) as temp_dirpath,
        temp.AutoTempFilePath() as temp_filepath,
    ):
      temp_linkpath = os.path.join(temp_dirpath, "foo")
      os.symlink(temp_filepath, temp_linkpath)

      stat = filesystem.Stat.FromPath(temp_linkpath, follow_symlink=False)
      self.assertTrue(stat.IsSymlink())
      self.assertEqual(stat.GetLinuxFlags(), 0)
      self.assertEqual(stat.GetOsxFlags(), 0)

  @unittest.skipIf(
      platform.system() == "Windows", "Windows does not support socket.AF_UNIX."
  )
  def testGetFlagsSocket(self):
    with temp.AutoTempDirPath(remove_non_empty=True) as temp_dirpath:
      temp_socketpath = os.path.join(temp_dirpath, "foo")

      # There is a limit on maximum length for a socket path [1]. Most of the
      # time, this should not be an issue (since generated paths are something
      # like `/tmp/tmppqnrQsZ/foo`, way below this limit). However, on strange
      # setups this might not always be the case. Since we don't want to fail
      # the test on such configurations, we simply skip it.
      #
      # pylint: disable=line-too-long
      # [1]: https://unix.stackexchange.com/questions/367008/why-is-socket-path-length-limited-to-a-hundred-chars
      # pylint: enable=ling-too-long
      if (platform.system() == "Linux" and len(temp_socketpath) > 108) or (
          platform.system() == "Darwin" and len(temp_socketpath) > 104
      ):
        message = "Generated path '{}' is too long for a socket path"
        self.skipTest(message.format(temp_socketpath))

      sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
      try:
        sock.bind(temp_socketpath)

        stat = filesystem.Stat.FromPath(temp_socketpath, follow_symlink=False)
        self.assertTrue(stat.IsSocket())
        self.assertEqual(stat.GetLinuxFlags(), 0)
        self.assertEqual(stat.GetOsxFlags(), 0)
      finally:
        sock.close()

  def _Touch(self, path, mode, date):
    fmt_date = date.strftime("%Y%m%d%H%M")
    subprocess.check_call(["touch", mode, "-t", fmt_date, path])

  @staticmethod
  def _EpochMicros(date):
    return int(date.strftime("%s")) * 1000000


class StatCacheTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    self.temp_dir = temp.TempDirPath()
    self.addCleanup(shutil.rmtree, self.temp_dir)

  def Path(self, *args):
    return os.path.join(self.temp_dir, *args)

  def testBasicUsage(self):
    with io.open(self.Path("foo"), "wb") as fd:
      fd.write(b"123")
    with io.open(self.Path("bar"), "wb") as fd:
      fd.write(b"123456")
    with io.open(self.Path("baz"), "wb") as fd:
      fd.write(b"123456789")

    stat_cache = filesystem.StatCache()

    with MockStat() as stat_mock:
      foo_stat = stat_cache.Get(self.Path("foo"))
      self.assertEqual(foo_stat.GetSize(), 3)
      self.assertTrue(stat_mock.FromPath.called)

    with MockStat() as stat_mock:
      bar_stat = stat_cache.Get(self.Path("bar"))
      self.assertEqual(bar_stat.GetSize(), 6)
      self.assertTrue(stat_mock.FromPath.called)

    with MockStat() as stat_mock:
      other_foo_stat = stat_cache.Get(self.Path("foo"))
      self.assertEqual(other_foo_stat.GetSize(), 3)
      self.assertFalse(stat_mock.FromPath.called)

    with MockStat() as stat_mock:
      other_bar_stat = stat_cache.Get(self.Path("bar"))
      self.assertEqual(other_bar_stat.GetSize(), 6)
      self.assertFalse(stat_mock.FromPath.called)

    with MockStat() as stat_mock:
      baz_stat = stat_cache.Get(self.Path("baz"))
      self.assertEqual(baz_stat.GetSize(), 9)
      self.assertTrue(stat_mock.FromPath.called)

    with MockStat() as stat_mock:
      other_baz_stat = stat_cache.Get(self.Path("baz"))
      self.assertEqual(other_baz_stat.GetSize(), 9)
      self.assertFalse(stat_mock.FromPath.called)

  @unittest.skipIf(
      platform.system() == "Windows", "Windows does not support os.symlink()."
  )
  def testFollowSymlink(self):
    with io.open(self.Path("foo"), "wb") as fd:
      fd.write(b"123456")
    os.symlink(self.Path("foo"), self.Path("bar"))

    stat_cache = filesystem.StatCache()

    with MockStat() as stat_mock:
      bar_stat = stat_cache.Get(self.Path("bar"), follow_symlink=False)
      self.assertTrue(bar_stat.IsSymlink())
      self.assertTrue(stat_mock.FromPath.called)

    with MockStat() as stat_mock:
      foo_stat = stat_cache.Get(self.Path("bar"), follow_symlink=True)
      self.assertFalse(foo_stat.IsSymlink())
      self.assertEqual(foo_stat.GetSize(), 6)
      self.assertTrue(stat_mock.FromPath.called)

  def testSmartSymlinkCache(self):
    with open(self.Path("foo"), "wb") as fd:
      fd.write(b"12345")

    stat_cache = filesystem.StatCache()

    with MockStat() as stat_mock:
      foo_stat = stat_cache.Get(self.Path("foo"), follow_symlink=False)
      self.assertEqual(foo_stat.GetSize(), 5)
      self.assertTrue(stat_mock.FromPath.called)

    with MockStat() as stat_mock:
      other_foo_stat = stat_cache.Get(self.Path("foo"), follow_symlink=True)
      self.assertEqual(other_foo_stat.GetSize(), 5)
      self.assertFalse(stat_mock.FromPath.called)


def MockStat():
  return mock.patch.object(filesystem, "Stat", wraps=filesystem.Stat)


if __name__ == "__main__":
  absltest.main()
